from biothings.utils.common import dotdict
import logging

class BiothingScrollError(Exception):
    ''' Error thrown when an ES scroll process errs '''
    pass

class BiothingSearchError(Exception):
    ''' Error thrown when given query errs (either from ES ``search_phase_exception``, or other errors). '''
    pass

class ESQuery(object):
    ''' This class contains functions to execute the *query* section of all handler pipelines.
    The inputs to it are an Elasticsearch client (from `BiothingESWebSettings`_), and any options
    from the URL string.  Each handler calls a different query function, though they all do essentially
    the same thing: get the query generated in the ESQueryBuilder stage of the pipeline (``query_kwargs``), and run it
    using the supplied Elasticsearch client.'''

    def __init__(self, client, options=dotdict()):
        self.client = client
        self.options = options

    async def _scroll(self, query_kwargs):
        ''' Returns the next scroll batch for the given scroll id '''
        from elasticsearch import NotFoundError, RequestError, TransportError
        try:
            return await self.client.scroll(**query_kwargs)
        except (NotFoundError, RequestError, TransportError):
            raise BiothingScrollError("Invalid or stale scroll_id")

    async def _annotation_GET_query(self, query_kwargs):
        if query_kwargs.get('id', None):
            # these query kwargs should be to an es.get
            return await self.get_biothing(query_kwargs)
        else:
            return await self.client.search(**query_kwargs)

    async def _annotation_POST_query(self, query_kwargs):
        return await self._common_POST_query(query_kwargs)

    async def _query_GET_query(self, query_kwargs, *args, **kwargs):
        from elasticsearch import RequestError
        try:
            return await self.client.search(**query_kwargs)
        except RequestError as e:
            if e.args[1] == 'search_phase_execution_exception' and "error" in e.args[2] and "root_cause" in e.args[2]["error"]:
                _root_causes = [
                    '{} {}'.format(c['type'],
                                   c['reason']) for c in e.args[2]['error']['root_cause']
                    if 'reason' in c and 'type' in c]
                raise BiothingSearchError(
                    'Could not execute query due to the following exception(s): {}'.format(_root_causes))
            else:
                raise Exception('{0}'.format(e))

    async def _query_POST_query(self, query_kwargs):
        return await self._common_POST_query(query_kwargs)

    async def _common_POST_query(self, query_kwargs):
        from elasticsearch import RequestError
        try:
            res = await self.client.msearch(**query_kwargs)
        except RequestError as e:
            if e.args[1] == 'search_phase_execution_exception' and "error" in e.args[2] and "root_cause" in e.args[2]["error"]:
                _root_causes = [
                    '{} {}'.format(c['type'],
                                   c['reason']) for c in e.args[2]['error']['root_cause']
                    if 'reason' in c and 'type' in c]
                raise BiothingSearchError(
                    'Could not execute query due to the following exception(s): {}'.format(_root_causes))
            else:
                raise Exception('{0}'.format(e))

        _root_causes = []
        for i in res['responses']:
            if 'error' in i:
                _root_causes.extend(['{} {}'.format(c['type'], c['reason'])
                                     for c in i['error']['root_cause']])

        if _root_causes:
            raise BiothingSearchError(
                'Could not execute query due to the following exception(s): {}'.format(_root_causes))

        return res

    async def _metadata_query(self, query_kwargs):
        return await self.client.indices.get_mapping(**query_kwargs)

    async def get_biothing(self, query_kwargs):
        ''' Return a biothing using the Elasticsearch client.get function '''
        from elasticsearch import NotFoundError
        try:
            return await self.client.get(**query_kwargs)
        except NotFoundError:
            return {}

    async def annotation_GET_query(self, query_kwargs):
        ''' Given ``query_kwargs`` from ESQueryBuilder, return results of annotation lookup GET query on ES client.'''
        return await self._annotation_GET_query(query_kwargs)

    async def annotation_POST_query(self, query_kwargs):
        ''' Given ``query_kwargs`` from ESQueryBuilder, return results of annotation lookup POST query on ES client.'''
        return await self._annotation_POST_query(query_kwargs)

    async def query_GET_query(self, query_kwargs, *args, **kwargs):
        ''' Given ``query_kwargs`` from ESQueryBuilder, return results of query GET on ES client.'''
        return await self._query_GET_query(query_kwargs)

    async def query_POST_query(self, query_kwargs, *args, **kwargs):
        ''' Given ``query_kwargs`` from ESQueryBuilder, return results of query POST on ES client.'''
        return await self._query_POST_query(query_kwargs)

    async def metadata_query(self, query_kwargs):
        ''' Given ``query_kwargs`` from ESQueryBuilder, return results of metadata query on ES client.'''
        return await self._metadata_query(query_kwargs)

    async def scroll(self, query_kwargs):
        ''' Given ``query_kwargs`` from ESQueryBuilder, return results of a scroll on ES client - returns next batch of results. '''
        return await self._scroll(query_kwargs)



from tornado.web import HTTPError
from biothings.web.api.es.handlers.base_handler import BaseESRequestHandler
from biothings.web.api.es.transform import ScrollIterationDone
from biothings.web.api.es.query import BiothingScrollError, BiothingSearchError
from biothings.web.api.helper import BiothingParameterTypeError
from biothings.utils.web import sum_arg_dicts
import logging

class QueryHandler(BaseESRequestHandler):
    ''' Request handlers for requests to the query endpoint '''

    def initialize(self, web_settings):
        ''' Tornado handler `.initialize() <http://www.tornadoweb.org/en/stable/web.html#tornado.web.RequestHandler.initialize>`_ function for all requests to the query endpoint.
        Here, the allowed arguments are set (depending on the request method) for each kwarg category.'''
        super(QueryHandler, self).initialize(web_settings)
        self.ga_event_object_ret['action'] = self.request.method
        if self.request.method == 'GET':
            self.ga_event_object_ret['action'] = self.web_settings.GA_ACTION_QUERY_GET
            self.control_kwargs = self.web_settings.QUERY_GET_CONTROL_KWARGS
            self.es_kwargs = self.web_settings.QUERY_GET_ES_KWARGS
            self.esqb_kwargs = self.web_settings.QUERY_GET_ESQB_KWARGS
            self.transform_kwargs = self.web_settings.QUERY_GET_TRANSFORM_KWARGS
        elif self.request.method == 'POST':
            self.ga_event_object_ret['action'] = self.web_settings.GA_ACTION_QUERY_POST
            self.control_kwargs = self.web_settings.QUERY_POST_CONTROL_KWARGS
            self.es_kwargs = self.web_settings.QUERY_POST_ES_KWARGS
            self.esqb_kwargs = self.web_settings.QUERY_POST_ESQB_KWARGS
            self.transform_kwargs = self.web_settings.QUERY_POST_TRANSFORM_KWARGS
        else:
            # handle other verbs?
            pass
        self.kwarg_settings = sum_arg_dicts(self.control_kwargs, self.es_kwargs,
                                            self.esqb_kwargs, self.transform_kwargs)
        logging.debug("QueryHandler - {}".format(self.request.method))
        logging.debug("Google Analytics Base object: {}".format(self.ga_event_object_ret))
        logging.debug("Kwarg Settings: {}".format(self.kwarg_settings))

    def _pre_scroll_transform_GET_hook(self, options, res):
        ''' Override me. '''
        return res

    async def get(self):
        ''' Handle a GET to the query endpoint. '''
        ###################################################
        #          Get/type/alias query parameters
        ###################################################

        try:
            kwargs = self.get_query_params()
        except BiothingParameterTypeError as e:
            self._return_data_and_track(
                {'success': False, 'error': "{0}".format(e)},
                ga_event_data={'total': 0},
                status_code=400)
            return
        #except Exception as e:
        #    self.log_exceptions("Error in get_query_params")
        #    self._return_data_and_track({'success': False, 'error': "Error parsing input parameter, check input types"}, ga_event_data={'total': 0})
        #    return

        ###################################################
        #      Split query parameters into categories
        ###################################################

        options = self.get_cleaned_options(kwargs)

        logging.debug("Request kwargs: {}".format(kwargs))
        logging.debug("Request options: {}".format(options))

        if not options.control_kwargs.q and not options.control_kwargs.scroll_id:
            self._return_data_and_track(
                {'success': False, 'error': "Missing required parameters."},
                ga_event_data={'total': 0},
                status_code=400, _format=options.control_kwargs.out_format)
            return

        options = self._pre_query_builder_GET_hook(options)

        ###################################################
        #          Instantiate pipeline classes
        ###################################################

        # Instantiate query builder, query, and transform classes
        _query_builder = self.web_settings.ES_QUERY_BUILDER(
            options=options.esqb_kwargs,
            index=self._get_es_index(options),
            doc_type=self._get_es_doc_type(options),
            es_options=options.es_kwargs,
            userquery_dir=self.web_settings.USERQUERY_DIR,
            scroll_options={
                'scroll': self.web_settings.ES_SCROLL_TIME,
                'size': self.web_settings.ES_SCROLL_SIZE},
            default_scopes=self.web_settings.DEFAULT_SCOPES,
            allow_random_query=self.web_settings.ALLOW_RANDOM_QUERY)
        _backend = self.web_settings.ES_QUERY(
            client=self.web_settings.async_es_client,
            options=options.es_kwargs)
        _result_transformer = self.web_settings.ES_RESULT_TRANSFORMER(
            options=options.transform_kwargs, host=self.request.host,
            jsonld_context=self.web_settings._jsonld_context,
            doc_url_function=self.web_settings.doc_url,
            output_aliases=self.web_settings.OUTPUT_KEY_ALIASES,
            source_metadata=self.web_settings.source_metadata(),
            licenses=self.web_settings.LICENSE_TRANSFORM)

        ###################################################
        #           Scroll request pipeline
        ###################################################

        if options.control_kwargs.scroll_id:
            ###################################################
            #             Build scroll query
            ###################################################

            _query = _query_builder.scroll(options.control_kwargs.scroll_id)
            #try:
            #    _query = _query_builder.scroll(options.control_kwargs.scroll_id)
            #except Exception as e:
            #    self.log_exceptions("Error building scroll query")
            #    self._return_data_and_track({'success': False, 'error': 'Error building scroll query for scroll_id "{}"'.format(options.control_kwargs.scroll_id)}, ga_event_data={'total': 0})
            #    return

            ###################################################
            #             Get scroll results
            ###################################################

            try:
                res = await _backend.scroll(_query)
            except BiothingScrollError as e:
                self._return_data_and_track(
                    {'success': False, 'error': '{}'.format(e)},
                    ga_event_data={'total': 0},
                    status_code=400, _format=options.control_kwargs.out_format)
                return
            #except Exception as e:
            #    self.log_exceptions("Error getting scroll batch")
            #    self._return_data_and_track({'success': False, 'error': 'Error retrieving scroll results for scroll_id "{}"'.format(options.control_kwargs.scroll_id)}, ga_event_data={'total': 0})
            #    return

            #logging.debug("Raw scroll query result: {}".format(res))

            if options.control_kwargs.raw:
                self._return_data_and_track(
                    res, ga_event_data={'total': res.get('total', 0)},
                    _format=options.control_kwargs.out_format)
                return

            res = self._pre_scroll_transform_GET_hook(options, res)

            ###################################################
            #             Transform scroll result
            ###################################################

            try:
                res = _result_transformer.clean_scroll_response(res)
            except ScrollIterationDone as e:
                self._return_data_and_track(
                    {'success': False, 'error': '{}'.format(e)},
                    ga_event_data={'total': res.get('total', 0)},
                    status_code=200, _format=options.control_kwargs.out_format)
                return
            #except Exception as e:
            #    self.log_exceptions("Error transforming scroll batch")
            #    self._return_data_and_track({'success': False, 'error': 'Error transforming scroll results for scroll_id "{}"'.format(options.control_kwargs.scroll_id)})
            #    return
        else:
            #####  Non-scroll query GET pipeline #############
            ###################################################
            #             Build query
            ###################################################

            _query = _query_builder.query_GET_query(q=options.control_kwargs.q)
            #try:
            #    _query = _query_builder.query_GET_query(q=options.control_kwargs.q)
            #except Exception as e:
            #    self.log_exceptions("Error building query")
            #    self._return_data_and_track({'success': False, 'error': 'Error building query from q="{}"'.format(options.control_kwargs.q)}, ga_event_object={'total': 0})
            #    return

            if options.control_kwargs.rawquery:
                self._return_data_and_track(
                    _query, ga_event_data={'total': 0},
                    rawquery=True, _format=options.control_kwargs.out_format)
                return

            _query = self._pre_query_GET_hook(options, _query)

            ###################################################
            #             Get query results
            ###################################################

            try:
                res = await _backend.query_GET_query(_query)
            except BiothingSearchError as e:
                self._return_data_and_track(
                    {'success': False, 'error': '{0}'.format(e)},
                    ga_event_data={'total': 0},
                    status_code=400, _format=options.control_kwargs.out_format)
                return
            #except Exception as e:
            #    self.log_exceptions("Error executing query")
            #    self._return_data_and_track({'success': False, 'error': 'Error executing query'},
            #                                    ga_event_data={'total': 0})
            #    return

            #logging.debug("Raw query result")
            #logging.debug("Raw query result: {}".format(res))

            # return raw result if requested
            if options.control_kwargs.raw:
                self._return_data_and_track(
                    res, ga_event_data={'total': res.get('total', 0)},
                    _format=options.control_kwargs.out_format)
                return

            res = self._pre_transform_GET_hook(options, res)

            ###################################################
            #            Transform query results
            ###################################################
            # clean result
            res = _result_transformer.clean_query_GET_response(res)
            #try:
            #    res = _result_transformer.clean_query_GET_response(res)
            #except Exception as e:
            #    self.log_exceptions("Error transforming query")
            #    logging.debug("Return query GET")
            #    self._return_data_and_track({'success': False, 'error': 'Error transforming query result'},
            #                                ga_event_data={'total': res.get('total', 0)})
            #    return

        res = self._pre_finish_GET_hook(options, res)

        # return and track
        self.return_object(res, _format=options.control_kwargs.out_format)
        if options.control_kwargs.fetch_all:
            self.ga_event_object_ret['action'] = 'fetch_all'
        self.ga_track(event=self.ga_event_object({'total': res.get('total', 0)}))
        self.self_track(data=self.ga_event_object_ret)
        return

    ###########################################################################

    async def post(self):
        ''' Handle a POST to the query endpoint.'''
        ###################################################
        #          Get/type/alias query parameters
        ###################################################

        try:
            kwargs = self.get_query_params()
        except BiothingParameterTypeError as e:
            self._return_data_and_track(
                {'success': False, 'error': "{0}".format(e)},
                ga_event_data={'qsize': 0},
                status_code=400)
            return
        #except Exception as e:
        #    self.log_exceptions("Error in get_query_params")
        #    self._return_data_and_track({'success': False, 'error': "Error parsing input parameter, check input types"}, ga_event_data={'qsize': 0})
        #    return

        options = self.get_cleaned_options(kwargs)

        logging.debug("Request kwargs: {}".format(kwargs))
        logging.debug("Request options: {}".format(options))

        if not options.control_kwargs.q:
            self._return_data_and_track(
                {'success': False, 'error': "Missing required parameters."},
                ga_event_data={'qsize': 0},
                status_code=400, _format=options.control_kwargs.out_format)
            return

        options = self._pre_query_builder_POST_hook(options)

        ###################################################
        #          Instantiate pipeline classes
        ###################################################

        # Instantiate query builder, query, and transform classes
        _query_builder = self.web_settings.ES_QUERY_BUILDER(
            options=options.esqb_kwargs,
            index=self._get_es_index(options),
            doc_type=self._get_es_doc_type(options),
            es_options=options.es_kwargs,
            userquery_dir=self.web_settings.USERQUERY_DIR,
            default_scopes=self.web_settings.DEFAULT_SCOPES)
        _backend = self.web_settings.ES_QUERY(
            client=self.web_settings.async_es_client,
            options=options.es_kwargs)
        _result_transformer = self.web_settings.ES_RESULT_TRANSFORMER(
            options=options.transform_kwargs, host=self.request.host,
            doc_url_function=self.web_settings.doc_url,
            jsonld_context=self.web_settings._jsonld_context,
            output_aliases=self.web_settings.OUTPUT_KEY_ALIASES,
            source_metadata=self.web_settings.source_metadata(),
            licenses=self.web_settings.LICENSE_TRANSFORM)

        ###################################################
        #                  Build query
        ###################################################

        _query = _query_builder.query_POST_query(
            qs=options.control_kwargs.q,
            scopes=options.esqb_kwargs.scopes)
        #try:
        #    _query = _query_builder.query_POST_query(qs=options.control_kwargs.q, scopes=options.esqb_kwargs.scopes)
        #except Exception as e:
        #    self.log_exceptions("Error building POST query")
        #    logging.debug("Returning query POST")
        #    self._return_data_and_track({'success': False, 'error': 'Error building query'}, ga_event_data={'qsize': len(options.control_kwargs.q)})
        #    return

        if options.control_kwargs.rawquery:
            self._return_data_and_track(
                _query,
                ga_event_data={
                    'qsize': len(
                        options.control_kwargs.q)},
                rawquery=True,
                _format=options.control_kwargs.out_format)
            return

        _query = self._pre_query_POST_hook(options, _query)

        ###################################################
        #                 Execute query
        ###################################################

        try:
            res = await _backend.query_POST_query(_query)
        except BiothingSearchError as e:
            self._return_data_and_track(
                {'success': False, 'error': '{0}'.format(e)},
                ga_event_data={'qsize': len(options.control_kwargs.q)},
                status_code=400, _format=options.control_kwargs.out_format)
            return
        #except Exception as e:
        #    self.log_exceptions("Error executing POST query")
        #    self._return_data_and_track({'success': False, 'error': 'Error executing query'}, ga_event_data={'qsize': len(options.control_kwargs.q)})
        #    return

        logging.debug("Raw query result: {}".format(res))

        # return raw result if requested
        if options.control_kwargs.raw:
            self._return_data_and_track(
                res, ga_event_data={'qsize': len(options.control_kwargs.q)},
                _format=options.control_kwargs.out_format)
            return

        res = self._pre_transform_POST_hook(options, res)

        ###################################################
        #          Transform query results
        ###################################################

        # clean result
        res = _result_transformer.clean_query_POST_response(qlist=options.control_kwargs.q, res=res)
        #try:
        #    res = _result_transformer.clean_query_POST_response(qlist=options.control_kwargs.q, res=res)
        #except Exception as e:
        #    self.log_exceptions("Error transforming POST query")
        #    self._return_data_and_track({'success': False, 'error': 'Error transforming query result'},
        #                        ga_event_data={'qsize': len(options.control_kwargs.q)})
        #    return

        res = self._pre_finish_POST_hook(options, res)

        # return and track
        self._return_data_and_track(
            res, ga_event_data={'qsize': len(options.control_kwargs.q)},
            _format=options.control_kwargs.out_format)



# -*- coding: utf-8 -*-
'''Settings objects used to configure the web API
These settings get passed into the handler.initialize() function,
of each request, and configure the web API endpoint.  They are mostly
a container for the `Config module`_, and any other settings that
are the same across all handler types, e.g. the Elasticsearch client.'''

import asyncio
import json
import logging
import os
import socket
import types
from importlib import import_module

from elasticsearch import Elasticsearch
from elasticsearch_async import AsyncElasticsearch


# Error class
class BiothingConfigError(Exception):
    pass

class BiothingWebSettings(object):
    ''' A container for the settings that configure the web API '''

    def __init__(self, config='biothings.web.settings.default'):
        ''' The ``config`` init parameter specifies a module that configures
        this biothing.  For more information see `config module`_ documentation.'''
        self.config_mod = isinstance(config, types.ModuleType) and config or import_module(config)
        try:
            with open(os.path.abspath(self.config_mod.JSONLD_CONTEXT_PATH), 'r') as json_file:
                self._jsonld_context = json.load(json_file)
        except BaseException:
            self._jsonld_context = {}

        # for metadata dev
        self._app_git_repo = os.path.abspath(
            self.APP_GIT_REPOSITORY) if hasattr(
            self, 'APP_GIT_REPOSITORY') else os.path.abspath('.')
        if not (self._app_git_repo and os.path.exists(self._app_git_repo) and os.path.isdir(
                self._app_git_repo) and os.path.exists(os.path.join(self._app_git_repo, '.git'))):
            self._app_git_repo = None

        # validate these settings?
        self.validate()

    def __getattr__(self, name):
        try:
            return getattr(self.config_mod, name)
        except AttributeError:
            raise AttributeError(
                "No setting named '{}' was found, check configuration module.".format(name))

    def set_debug_level(self, debug=False):
        '''Set if running API in debug mode.
        Should be called before passing ``self`` to handler initialization.'''
        self._DEBUG = debug
        return self

    def generate_app_list(self):
        ''' Generates the tornado.web.Application `(regex, handler_class, options) tuples <http://www.tornadoweb.org/en/stable/web.html#application-configuration>`_ for this project.'''
        return self.UNINITIALIZED_APP_LIST + [
            (endpoint_regex, handler, {"web_settings": self})
            for (endpoint_regex, handler) in self.APP_LIST]

    def validate(self):
        ''' Validate these settings '''
        pass

class BiothingESWebSettings(BiothingWebSettings):
    ''' `BiothingWebSettings`_ subclass with functions specific to an elasticsearch backend '''

    def __init__(self, config='biothings.web.settings.default'):
        ''' The ``config`` init parameter specifies a module that configures
        this biothing.  For more information see `config module`_ documentation.'''
        super(BiothingESWebSettings, self).__init__(config)

        # get es client for web
        self.es_client = self.get_es_client()
        self.async_es_client = self.get_es_client(sync=False)

        # populate the metadata for this project
        self.source_metadata()

        # initialize payload for standalone tracking batch
        self.tracking_payload = []

    def doc_url(self, bid):
        ''' Function to return a url on this biothing API to the biothing object specified by bid.'''
        return os.path.join(self.URL_BASE, self.API_VERSION, self.ES_DOC_TYPE, bid)

    def _source_metadata_object(self):
        ''' Override me to return metadata for your project '''
        _meta = {}
        try:
            _m = self.es_client.indices.get_mapping(index=self.ES_INDEX, doc_type=self.ES_DOC_TYPE)
            _meta = _m[list(_m.keys())[0]]['mappings'][self.ES_DOC_TYPE]['_meta']['src']
        except BaseException:
            pass
        return _meta

    def source_metadata(self):
        ''' Function to cache return of the source metadata stored in _meta of index mappings '''
        if getattr(self, '_source_metadata', False):
            return self._source_metadata

        self._source_metadata = self._source_metadata_object()

        return self._source_metadata

    def available_fields_notes(self):
        ''' Caches the available fields notes for this biothing '''
        try:
            return self._available_fields_notes
        except BaseException:
            pass

        self._available_fields_notes = {}

        if os.path.exists(os.path.abspath(self.AVAILABLE_FIELDS_NOTES_PATH)):
            try:
                with open(os.path.abspath(self.AVAILABLE_FIELDS_NOTES_PATH), 'r') as inf:
                    self._available_fields_notes = json.load(inf)
            except BaseException:
                pass

        return self._available_fields_notes

    def get_es_client(self, sync=True):
        '''
        Get the `Elasticsearch client <https://elasticsearch-py.readthedocs.io/en/master/>`_
        for this app, only called once on invocation of server.
        '''

        if sync:
            return Elasticsearch(self.ES_HOST, timeout=getattr(self, 'ES_CLIENT_TIMEOUT', 120))

        return AsyncElasticsearch(self.ES_HOST, timeout=getattr(self, 'ES_CLIENT_TIMEOUT', 120))



import os, sys, types
import asyncio
import logging
from pprint import pformat
import importlib

from biothings.utils.hub_db import get_data_plugin
from biothings.utils.dataload import to_boolean
from biothings.utils.manager import BaseSourceManager
from biothings.utils.hub_db import get_src_master, get_source_fullname, \
                                   get_src_dump


class SourceManager(BaseSourceManager):
    """
    Helper class to get information about a datasource,
    whether it has a dumper and/or uploaders associated.
    """

    def __init__(self, source_list, dump_manager, upload_manager, data_plugin_manager):
        self._orig_source_list = source_list
        self.source_list = None
        self.dump_manager = dump_manager
        self.upload_manager = upload_manager
        self.data_plugin_manager = data_plugin_manager
        self.reload()
        self.src_master = get_src_master()
        self.src_dump = get_src_dump()
        # honoring BaseSourceManager interface (gloups...-
        self.register = {}

    def reload(self):
        # clear registers
        self.dump_manager.register.clear()
        self.upload_manager.register.clear()
        # re-eval source list (so if it's a string, it'll re-discover sources)
        self.source_list = self.find_sources(self._orig_source_list)
        self.dump_manager.register_sources(self.source_list)
        self.upload_manager.register_sources(self.source_list)

    def find_sources(self,paths):
        sources = []

        if not type(paths) == list:
            paths = [paths]

        def eval_one_source(one_path):
            if "/" in one_path:
                # it's path to directory
                # expecting 
                if not one_path in sys.path:
                    logging.info("Adding '%s' to python path" % one_path)
                    sys.path.insert(0,one_path)
                for d in os.listdir(one_path):
                    if d.endswith("__pycache__"):
                        continue
                    sources.append(d)
            else:
                # assuming it's path to a python module (oath.to.module)
                sources.append(one_path)

        def eval_one_root(root):
            logging.debug("Discovering sources in %s" % root)
            # root is a module path where sources can be found
            rootdir,__init__ = os.path.split(root.__file__)
            for srcdir in os.listdir(rootdir):
                if srcdir.endswith("__pycache__"):
                    continue
                srcpath = os.path.join(rootdir,srcdir)
                if os.path.isdir(srcpath):
                    srcmod_str = "%s.%s" % (root.__name__,srcdir)
                    sources.append(srcmod_str)

        for path in paths:
            if type(path) == str:
                eval_one_source(path)
            elif type(path) == types.ModuleType:
                eval_one_root(path)

        # clean with only those which can be imported
        sources = set(sources)
        for s in [s for s in sources]:
            try:
                importlib.import_module(s)
            except Exception as e:
                logging.error("Failed to discover source '%s': %s" % (s,e))
                sources.remove(s)

        logging.info("Found sources: %s" % sorted(sources))
        return sources

    def set_mapping_src_meta(self, subsrc, mini):
        # get mapping from uploader klass first (hard-coded), then src_master (generated/manual)
        src_meta = {}
        mapping = {}
        origin = None
        try:
            upk = self.upload_manager["%s.%s" % (mini["_id"],subsrc)]
            assert len(upk) == 1, "More than 1 uploader found, can't handle that..."
            upk = upk.pop()
            src_meta = upk.__metadata__["src_meta"]
            mapping = upk.get_mapping()
            origin = "uploader"
            if not mapping:
                raise AttributeError("Not hard-coded mapping")
        except (IndexError, KeyError, AttributeError) as e:
            logging.debug("Can't find hard-coded mapping, now searching src_master: %s" % e)
            m = self.src_master.find_one({"_id":subsrc})
            mapping = m and m.get("mapping")
            origin = "master"
            # use metadata from upload or reconstitute(-ish)
            src_meta = src_meta or \
                    m and dict([(k,v) for (k,v) in m.items() if not k in ["_id","name","timestamp","mapping"]])
        if mapping:
            mini.setdefault("mapping",{}).setdefault(subsrc,{}).setdefault("mapping",mapping)
            mini.setdefault("mapping",{}).setdefault(subsrc,{}).setdefault("origin",origin)
        if src_meta:
            mini.setdefault("__metadata__",{}).setdefault(subsrc,src_meta)

    def sumup_source(self,src,detailed=False):
        """Return minimal info about src"""

        mini = {}
        mini["_id"] = src.get("_id",src["name"])
        mini["name"] = src["name"]

        if src.get("download"):
            mini["download"] = {
                    "status" : src["download"].get("status"),
                    "time" : src["download"].get("time"),
                    "started_at" : src["download"].get("started_at"),
                    "release" : src["download"].get("release"),
                    "data_folder" : src["download"].get("data_folder"),
                    }
            mini["download"]["dumper"] = src["download"].get("dumper",{})
            if src["download"].get("err"):
                mini["download"]["error"] = src["download"]["err"]

        count = 0
        if src.get("upload"):
            mini["upload"] = {"sources" : {}}
            for job,info in src["upload"]["jobs"].items():
                mini["upload"]["sources"][job] = {
                        "time" : info.get("time"),
                        "status" : info.get("status"),
                        "count" : info.get("count"),
                        "started_at" : info.get("started_at"),
                        "release" : info.get("release"),
                        "data_folder" : info.get("data_folder"),
                        }
                if info.get("err"):
                     mini["upload"]["sources"][job]["error"] = info["err"]
                count += info.get("count") or 0
                if detailed:
                    self.set_mapping_src_meta(job,mini)
        if src.get("inspect"):
            mini["inspect"] = {"sources" : {}}
            for job,info in src["inspect"]["jobs"].items():
                if not detailed:
                    # remove big inspect data but preserve inspect status/info and errors
                    mode_has_error = []
                    mode_ok = []
                    for mode in info.get("inspect",{}).get("results",{}):
                        if info["inspect"]["results"][mode].get("errors"):
                            mode_has_error.append(mode)
                        else:
                            mode_ok.append(mode)
                    for mode in mode_ok:
                        info["inspect"]["results"].pop(mode)
                    for mode in mode_has_error:
                        keys = list(info["inspect"]["results"][mode].keys())
                        # remove all except errors
                        for k in keys:
                            if k != "errors":
                                info["inspect"]["results"][mode].pop(k)
                        
                mini["inspect"]["sources"][job] = info

        if src.get("locked"):
            mini["locked"] = src["locked"]
        mini["count"] = count

        return mini

    def get_sources(self,id=None,debug=False,detailed=False):
        dm = self.dump_manager
        um = self.upload_manager
        dpm = self.data_plugin_manager
        ids = set()
        if id and id in dm.register:
            ids.add(id)
        elif id and id in um.register:
            ids.add(id)
        elif id and id in dpm.register:
            ids.add(id)
        else:
            # either no id passed, or doesn't exist
            if id and not len(ids):
                raise ValueError("Source %s doesn't exist" % repr(id))
            ids = set(dm.register)
            ids.update(um.register)
            ids.update(dpm.register)
        sources = {}
        bydsrcs = {}
        byusrcs = {}
        bydpsrcs = {}
        plugins = get_data_plugin().find()
        [bydsrcs.setdefault(src["_id"],src) for src in dm.source_info() if dm]
        [byusrcs.setdefault(src["_id"],src) for src in um.source_info() if um]
        [bydpsrcs.setdefault(src["_id"],src) for src in plugins]
        for _id in ids:
            # start with dumper info
            if dm:
                src = bydsrcs.get(_id)
                if src:
                    if debug:
                        sources[src["name"]] = src
                    else:
                        sources[src["name"]] = self.sumup_source(src,detailed)
            # complete with uploader info
            if um:
                src = byusrcs.get(_id)
                if src:
                    # collection-only source don't have dumpers and only exist in
                    # the uploader manager
                    if not src["_id"] in sources:
                        sources[src["_id"]] = self.sumup_source(src,detailed)
                    if src.get("upload"):
                        for subname in src["upload"].get("jobs",{}):
                            try:
                                sources[src["name"]].setdefault("upload",{"sources" : {}})["sources"].setdefault(subname,{})
                                sources[src["name"]]["upload"]["sources"][subname]["uploader"] = src["upload"]["jobs"][subname].get("uploader")
                            except Exception as e:
                                logging.error("Source is invalid: %s\n%s" % (e,pformat(src)))
            # deal with plugin info if any
            if dpm:
                src = bydpsrcs.get(_id)
                if src:
                    assert len(dpm[_id]) == 1, "Expected only one uploader, got: %s" % dpm[_id]
                    klass = dpm[_id][0]
                    src.pop("_id")
                    if hasattr(klass,"data_plugin_error"):
                        src["error"] = klass.data_plugin_error
                    sources.setdefault(_id,{"data_plugin": {}})
                    if src.get("download",{}).get("err"):
                        src["download"]["error"] = src["download"].pop("err")
                    sources[_id]["data_plugin"] = src
                    sources[_id]["_id"] = _id
                    sources[_id]["name"] = _id
        if id:
            src = list(sources.values()).pop()
            # enrich with metadata (uploader > dumper)
            ks = []
            if dm:
                try:
                    ks.extend(dm.register[id])
                except KeyError:
                    pass
            if um:
                try:
                    ks.extend(um.register[id])
                except KeyError:
                    pass
            for upk in ks:
                # name either from uploader or dumper
                name = getattr(upk,"name",None) or upk.SRC_NAME
                if getattr(upk,"__metadata__",{}).get("src_meta"):
                    src.setdefault("__metadata__",{}).setdefault(name,{})
                    src["__metadata__"][name] = upk.__metadata__["src_meta"]
            # simplify as needed (if only one source in metadata, remove source key level,
            # or if licenses are the same amongst sources, keep one copy)
            if len(src.get("__metadata__",{})) == 1:
                src["__metadata__"] = list(src["__metadata__"].values()).pop()
            elif len(src.get("__metadata__",{})) > 1:
                metas = list(src["__metadata__"].values())
                simplified = [metas.pop()]
                same = True
                while metas:
                    m = metas.pop()
                    if not m in simplified:
                        same = False
                        break
                if same:
                    # we consume all of them, ie. they're all equals
                    src["__metadata__"] = list(src["__metadata__"].values()).pop()
                else:
                    # convert to a list of dict (so it's easier to detect if one or more
                    # licenses just by checking if type is dict (one) or array (more))
                    metas = src.pop("__metadata__")
                    src["__metadata__"] = []
                    for m in metas:
                        src["__metadata__"].append({m:metas[m]})
            return src
        else:
            return list(sources.values())

    def get_source(self,name,debug=False):
        return self.get_sources(id=name,debug=debug,detailed=True)

    def save_mapping(self, name, mapping=None, dest="master", mode="mapping"):
        logging.debug("Saving mapping for source '%s' destination='%s':\n%s" % (name,dest,pformat(mapping)))
        # either given a fully qualified source or just sub-source
        try:
            subsrc = name.split(".")[1]
        except IndexError:
            subsrc = name
        if dest == "master":
            m = self.src_master.find_one({"_id":subsrc}) or {"_id":subsrc}
            m["mapping"] = mapping
            self.src_master.save(m)
        elif dest == "inspect":
            m = self.src_dump.find_one({"_id":name})
            try:
                m["inspect"]["jobs"][subsrc]["inspect"]["results"][mode] = mapping
                self.src_dump.save(m)
            except KeyError as e:
                raise ValueError("Can't save mapping, document doesn't contain expected inspection data" % e)
        else:
            raise ValueError("Unknow saving destination: %s" % repr(dest))




import os, time, datetime
import logging
import asyncio
from functools import partial

from .slack import slack_msg

def setup_default_log(default_logger_name,log_folder,level=logging.DEBUG):
    # this will affect any logging calls
    logging.basicConfig(level=level)
    if not os.path.exists(log_folder):
        os.makedirs(log_folder)
    logfile = os.path.join(log_folder, '%s_%s_hub.log' % (default_logger_name,time.strftime("%Y%m%d",datetime.datetime.now().timetuple())))
    fh = logging.FileHandler(logfile)
    fh.setFormatter(logging.Formatter('%(asctime)s [%(process)d:%(threadName)s] - %(name)s - %(levelname)s -- %(message)s',datefmt="%H:%M:%S"))
    fh.name = "logfile"
    logger = logging.getLogger(default_logger_name)
    logger.setLevel(level)
    if not fh.name in [h.name for h in logger.handlers]:
        logger.addHandler(fh)
    return logger


def get_logger(logger_name,log_folder=None,handlers=["console","file","slack"],timestamp="%Y%m%d"):
    """
    Configure a logger object from logger_name and return (logger, logfile)
    """
    from biothings import config as btconfig
    if not log_folder:
        log_folder = btconfig.LOG_FOLDER
    # this will affect any logging calls
    if not os.path.exists(log_folder):
        os.makedirs(log_folder)
    if timestamp:
        logfile = os.path.join(log_folder, '%s_%s.log' % (logger_name,time.strftime(timestamp,datetime.datetime.now().timetuple())))
    else:
        logfile = os.path.join(log_folder, '%s.log' % logger_name)
    fmt = logging.Formatter('%(asctime)s [%(process)d:%(threadName)s] - %(name)s - %(levelname)s -- %(message)s', datefmt="%H:%M:%S")
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    if "file" in handlers:
        fh = logging.FileHandler(logfile)
        fh.setFormatter(fmt)
        fh.name = "logfile"
        if not fh.name in [h.name for h in logger.handlers]:
            logger.addHandler(fh)

    if "hipchat" in handlers:
        raise DeprecationWarning("Hipchat is dead...")

    if "slack" in handlers and getattr(btconfig,"SLACK_WEBHOOK",None):
        nh = SlackHandler(btconfig.SLACK_WEBHOOK)
        nh.setFormatter(fmt)
        nh.name = "slack"
        if not nh.name in [h.name for h in logger.handlers]:
            logger.addHandler(nh)

    return (logger, logfile)


class SlackHandler(logging.StreamHandler):

    colors = {logging.DEBUG : "#a1a1a1",
              logging.INFO : "good",
              logging.WARNING : "warning",
              logging.ERROR : "danger",
              logging.CRITICAL : "#7b0099"}

    def __init__(self,webhook):
        super(SlackHandler,self).__init__()
        self.webhook = webhook

    def emit(self,record):
        @asyncio.coroutine
        def aioemit():
            fut = yield from loop.run_in_executor(None,partial(
                slack_msg,self.webhook,msg,color=color))
            return fut
        if record.__dict__.get("notify"):
            try:
                loop = asyncio.get_event_loop()
            except RuntimeError:
                # if the logger is running in a thread, there's no asyncio loop there
                # (we usually take it from job_manager.loop, but it's not accessible there)
                # so we use another loop
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
            msg = self.format(record)
            color = self.__class__.colors.get(record.levelno,self.__class__.colors[logging.DEBUG])
            fut = aioemit()
            asyncio.ensure_future(fut)


class EventRecorder(logging.StreamHandler):

    def __init__(self, *args, **kwargs):
        super(EventRecorder,self).__init__(*args,**kwargs)
        from biothings.utils.hub_db import get_event
        self.eventcol = get_event()

    def emit(self,record):
        @asyncio.coroutine
        def aioemit(msg):
            def recorded(f):
                res = f.result()
            fut = loop.run_in_executor(None,
                    partial(self.eventcol.save,msg))
            fut.add_done_callback(recorded)
            yield from fut
            return fut
        if record.__dict__.get("notify") or record.__dict__.get("event"):
            try:
                loop = asyncio.get_event_loop()
                msg = {
                        "_id" : record.created,
                        "asctime" : record.asctime,
                        "msg" : record.message,
                        "level" : record.levelname,
                        "name" : record.name,
                        "pid" : record.process,
                        "pname" : record.processName,
                        }
                fut = aioemit(msg)
                asyncio.ensure_future(fut)
            except Exception as e:
                logging.error("Couldn't record event: %s" % e)


class WSLogHandler(logging.StreamHandler):
    """
    when listener is a bt.hub.api.handlers.ws.LogListener instance,
    log statements are propagated through existing websocket
    """

    def __init__(self,listener):
        super().__init__()
        self.listener = listener
        self.count = 0

    def payload(self, record):
        msg = self.format(record)
        return {
                "_id" : self.count,
                "op" : "log",
                "msg" : msg,
                "logger" : record.name,
                "level" : record.levelname,
                "ts" : datetime.datetime.now().isoformat(),
                }

    def emit(self,record):
        self.count += 1
        self.listener.read(self.payload(record))

class WSShellHandler(WSLogHandler):
    """
    when listener is a bt.hub.api.handlers.ws.LogListener instance,
    log statements are propagated through existing websocket
    """

    def payload(self, record):
        types = {
            ShellLogger.INPUT : "input",
            ShellLogger.OUTPUT : "output"
        }
        return  {
                "_id" : self.count,
                "op" : "shell",
                "cmd" : record.msg,
                "type" : types.get(record.levelno,"unknown"),
                "ts" : datetime.datetime.now().isoformat(),
                }


class ShellLogger(logging.Logger):
    """
    Custom "levels" for input going to the shell
    and output coming from it (just for naming)
    """

    OUTPUT = 1000 
    INPUT = 1001

    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self.manager.loggerDict[self.name] = self

    def input(self, msg, *args, **kwargs):
        self._log(self.__class__.INPUT, msg, args, **kwargs)

    def output(self, msg, *args, **kwargs):
        self._log(self.__class__.OUTPUT, msg, args, **kwargs)




'''BioThings API ioloop start utilities.

This module contains functions to configure and start the `base event loop <http://www.tornadoweb.org/en/stable/ioloop.html>`_ from command line args.  Command line processing is done using tornado.options, with the following arguments defined:

    * ``port``: the port to start the API on, **default** 8000
    * ``address``: the address to start the API on, **default** 127.0.0.1
    * ``debug``: start the API in debug mode, **default** False
    * ``appdir``: path to API configuration directory, **default**: current working directory

    The **main** function is the boot script for all BioThings API webservers.
'''
import sys
import os
import tornado.httpserver
import tornado.ioloop
import tornado.web
import tornado.escape
from tornado.options import define, options

__USE_SENTRY__ = True
try:
    from raven.contrib.tornado import AsyncSentryClient
except ImportError:
    __USE_SENTRY__ = False
    def AsyncSentryClient(*args, **kwargs):
        pass

__USE_WSGI__ = False

define("port", default=8000, help="run on the given port", type=int)
define("address", default="127.0.0.1", help="run on localhost")
define("debug", default=False, type=bool, help="run in debug mode")
define("appdir", default=os.getcwd(), type=str, help="path to app directory containing (at minimum) a config module")

try:
    options.parse_command_line()
except:
    pass

# assume config file is root of appdir
src_path = os.path.abspath(options.appdir)
if src_path not in sys.path:
    sys.path.append(src_path)

if options.debug:
    import tornado.autoreload
    import logging
    logging.getLogger().setLevel(logging.DEBUG)
    options.address = '0.0.0.0'

def get_app(APP_LIST, **settings):
    ''' Return an Application instance. '''
    return tornado.web.Application(APP_LIST, **settings)

def main(APP_LIST, app_settings={}, debug_settings={}, sentry_client_key=None, use_curl=False):
    ''' Main ioloop configuration and start

        :param APP_LIST: a list of `URLSpec objects or (regex, handler_class) tuples <http://www.tornadoweb.org/en/stable/web.html#tornado.web.Application>`_
        :param app_settings: `Tornado application settings <http://www.tornadoweb.org/en/stable/web.html#tornado.web.Application.settings>`_
        :param debug_settings: Additional application settings for API debug mode
        :param sentry_client_key: Application-specific key for attaching Sentry monitor to the application
        :param use_curl: Overide the default simple_httpclient with curl_httpclient (Useful for Github Login) <https://www.tornadoweb.org/en/stable/httpclient.html>
    '''
    settings = app_settings
    if options.debug:
        settings.update(debug_settings)
        settings.update({"debug": True})
    application = get_app(APP_LIST, **settings)
    if __USE_SENTRY__ and sentry_client_key:
       application.sentry_client = AsyncSentryClient(sentry_client_key)
    if use_curl:
        tornado.httpclient.AsyncHTTPClient.configure(
            "tornado.curl_httpclient.CurlAsyncHTTPClient")
    http_server = tornado.httpserver.HTTPServer(application)
    http_server.listen(options.port, address=options.address)
    loop = tornado.ioloop.IOLoop.instance()
    if options.debug:
        tornado.autoreload.start(loop)
        logging.info('Server is running on "%s:%s"...' % (options.address, options.port))
    loop.start()



import sys, re, os, time, math, glob, copy
from datetime import datetime
from dateutil.parser import parse as dtparse
import pickle, json
from pprint import pformat
import asyncio
from functools import partial
from elasticsearch import Elasticsearch

import biothings.utils.mongo as mongo
from biothings.utils.hub_db import get_src_build
import biothings.utils.aws as aws
from biothings.utils.common import timesofar, get_random_string, iter_n, \
                                   get_class_from_classpath, get_dotfield_value
from biothings.utils.loggers import get_logger
from biothings.utils.manager import BaseManager, BaseStatusRegisterer
from biothings.utils.es import ESIndexer, IndexerException as ESIndexerException
from biothings.utils.backend import DocESBackend
from biothings import config as btconfig
from biothings.utils.mongo import doc_feeder, id_feeder
from config import LOG_FOLDER, logger as logging
from biothings.utils.hub import publish_data_version, template_out
from biothings.hub.databuild.backend import generate_folder, create_backend, \
                                            merge_src_build_metadata
from biothings.hub import SNAPSHOOTER_CATEGORY, SNAPSHOTMANAGER_CATEGORY


class Snapshooter(BaseStatusRegisterer):

    def __init__(self, job_manager, index_manager, envconf, log_folder, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index_manager = index_manager
        self.job_manager = job_manager
        self.envconf = envconf
        self.log_folder = log_folder
        self.ti = time.time()
        self.setup()
        # extract some "special" values depending on the config type, etc...
        if self.envconf.get("repository",{}).get("type") == "fs":
            self.es_backups_folder = self.envconf["repository"]["es_backups_folder"]

    @property
    def collection(self):
        return get_src_build()

    def setup(self):
        self.setup_log()
    
    def setup_log(self):
        self.logger, self.logfile = get_logger(SNAPSHOOTER_CATEGORY,self.log_folder)

    def get_predicates(self):
        return []
    
    def get_pinfo(self):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        pinfo = {"category" : SNAPSHOOTER_CATEGORY,
                "source" : "",
                "step" : "",
                "description" : ""}
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    def load_build(self, index_name):
        return self.load_doc(index_name,"index")

    def template_out_conf(self, build_doc):
        """
        Template out for special value using build_doc
        Templated values can look like:
            "base_path" : "onefolder/%(_meta.build_version)s"
        where "_meta.build_version" value is taken from build_doc dictionary
        (dot field notation).  In other words, such repo config are dynamic
        and potentially change for each index/snapshot created.
        """
        try:
            strconf = template_out(json.dumps(self.envconf),build_doc)
            return json.loads(strconf)
        except Exception as e:
            self.logger.exception("Coudn't template out configuration: %s" % e)
            raise

    def reset_repository_info(self, bdoc, snapshot_name):
        if bdoc.get("snapshot",{}).get(snapshot_name):
            if bdoc["snapshot"][snapshot_name].pop("repository",None):
                    self.collection.save(bdoc)

    def register_status(self, bdoc, status,transient=False,init=False,**extra):
        super().register_status(bdoc,"snapshot",status,transient=transient,init=init,**extra)

    def get_es_idxr(self, envconf, index=None):
        if envconf["indexer"].get("env"):
            # we should take indexer params from INDEX_CONFIG, ie. index_manager
            idxklass = self.index_manager.find_indexer(index)
            idxkwargs = self.index_manager[envconf["indexer"]["env"]]
        else:
            idxklass = self.index_manager.DEFAULT_INDEXER
            es_host = envconf["indexer"]["host"]
            idxkwargs = envconf["indexer"]["host"]["args"]
        idxr = idxklass(**idxkwargs)
        es_idxr = ESIndexer(index=index,doc_type=idxr.doc_type,
                            es_host=idxr.host,
                            check_index=not index is None)

        return es_idxr

    def pre_snapshot(self, envconf, repo_conf, index_meta):
        pass

    def snapshot(self, index, snapshot=None, steps=["pre","snapshot","post"]):
        bdoc = self.load_build(index)
        self.envconf = self.template_out_conf(bdoc)
        # check what to do
        if type(steps) == str:
            steps = [steps]
        snapshot_name = snapshot or index
        es_idxr = self.get_es_idxr(self.envconf,index)
        # create repo if needed
        index_meta = es_idxr.get_mapping_meta()["_meta"] # read from index
        repo_name, repo_conf = self.create_repository(self.envconf, bdoc, index_meta)
        monitor_delay = self.envconf["monitor_delay"]
        # will hold the overall result
        fut = asyncio.Future()

        def get_status():
            try:
                res = es_idxr.get_snapshot_status(repo_name, snapshot_name)
                assert "snapshots" in res, "Can't find snapshot '%s' in repository '%s'" % (snapshot_name,repo_name)
                # assuming only one index in the snapshot, so only check first elem
                info = res["snapshots"][0]
                assert info.get("state"), "Can't find state in snapshot '%s'" % snapshot_name
                return info
            except Exception as e:
                # somethng went wrong, report as failure
                return "FAILED"

        @asyncio.coroutine
        def do(index):
            try:
                global_state = None
                got_error = None

                def snapshot_launched(f):
                    try:
                        self.logger.info("Snapshot launched: %s" % f.result())
                    except Exception as e:
                        self.logger.error("Error while lauching snapshot: %s" % e)
                        nonlocal got_error
                        got_error = e
                
                def done(f,step):
                    try:
                        res = f.result()
                        self.register_status(bdoc,"success",
                                job={"step":"%s-snapshot" % step,"result" : res},
                                snapshot={snapshot_name : {"conf" : self.envconf,step : res}})
                        self.logger.info("%s-snapshot done: %s" % (step,res))
                    except Exception as e:
                        nonlocal got_error
                        got_error = e
                        self.register_status(bdoc,"failed",
                                job={"step":"%s-snapshot" % step,"err" : str(e)},
                                snapshot={snapshot_name : {"conf" : self.envconf,step : None}})
                        self.logger.exception("Error while running pre-snapshot: %s" % e)

                pinfo = self.get_pinfo()
                pinfo["source"] = index

                # we only allow one repo conf per snapshot name
                self.reset_repository_info(bdoc,snapshot_name)

                if "pre" in steps:
                    self.register_status(bdoc,"pre-snapshotting",transient=True,init=True,
                                         job={"step":"pre-snapshot"},snapshot={snapshot_name:{}})
                    pinfo["step"] = "pre-snapshot"
                    pinfo.pop("description",None)
                    job = yield from self.job_manager.defer_to_thread(pinfo,
                            partial(self.pre_snapshot,self.envconf,repo_conf,index_meta))
                    job.add_done_callback(partial(done,step="pre"))
                    yield from job
                    if got_error:
                        fut.set_exception(got_error)
                        return
                    
                if "snapshot" in steps:
                    self.register_status(bdoc,"snapshotting",transient=True,init=True,
                                         job={"step":"snapshot"},snapshot={snapshot_name:{}})
                    pinfo["step"] = "snapshot"
                    pinfo["description"] = es_idxr.es_host
                    self.logger.info("Creating snapshot for index '%s' on host '%s', repository '%s'" % (index,es_idxr.es_host,repo_name))
                    job = yield from self.job_manager.defer_to_thread(pinfo,
                            partial(es_idxr.snapshot,repo_name,snapshot_name))
                    job.add_done_callback(snapshot_launched)
                    yield from job

                    # launched, so now monitor completion
                    while True:
                        info = get_status()
                        state = info["state"]
                        failed_shards = info["shards_stats"]["failed"]
                        if state in ["INIT","IN_PROGRESS","STARTED"]:
                            yield from asyncio.sleep(monitor_delay)
                        else:
                            if state == "SUCCESS" and failed_shards == 0:
                                global_state = state.lower()
                                self.logger.info("Snapshot '%s' successfully created (host: '%s', repository: '%s')" % \
                                        (snapshot_name,es_idxr.es_host,repo_name),extra={"notify":True})
                            else:
                                if state == "SUCCESS":
                                    e = IndexerException("Snapshot '%s' partially failed: state is %s but %s shards failed" % (snapshot_name,state,failed_shards))
                                else:
                                    e = IndexerException("Snapshot '%s' failed: state is %s" % (snapshot_name,state))
                                self.logger.error("Failed creating snapshot '%s' (host: %s, repository: %s), state: %s" % \
                                        (snapshot_name,es_idxr.es_host,repo_name,state),extra={"notify":True})
                                got_error = e
                            break

                    if got_error:
                        self.register_status(
                                bdoc,
                                "failed",
                                job={
                                    "step":"snapshot",
                                    "err" : str(got_error)
                                    },
                                snapshot={
                                    snapshot_name : {
                                        "conf" : self.envconf,
                                        "repository": None,
                                        }
                                    }
                                )
                        fut.set_exception(got_error)
                        return

                    else:
                        self.register_status(
                                bdoc,
                                "success",
                                job={
                                    "step" : "snapshot",
                                    "status" : global_state,
                                    },
                                snapshot={
                                    snapshot_name : {
                                        "conf" : self.envconf,
                                        "repository": {repo_name : repo_conf}
                                        }
                                    }
                                )

                if "post" in steps:
                    self.register_status(bdoc,"post-snapshotting",transient=True,init=True,
                                         job={"step":"post-snapshot"},snapshot={snapshot_name:{}})
                    pinfo["step"] = "post-snapshot"
                    pinfo.pop("description",None)
                    job = yield from self.job_manager.defer_to_thread(pinfo,
                            partial(self.post_snapshot,self.envconf,repo_conf,index_meta))
                    job.add_done_callback(partial(done,step="post"))
                    yield from job
                    if got_error:
                        fut.set_exception(got_error)
                        return

                # if  we get there all is fine
                fut.set_result(global_state)

            except Exception as e:
                self.logger.exception(e)
                fut.set_exception(e)

        task = asyncio.ensure_future(do(index))
        return fut

    def post_snapshot(self, envconf, repo_conf, index_meta):
        pass

    def get_repository_config(self, repo_conf, build_doc, index_meta):
        repo_name = repo_conf["name"]
        repo_type = repo_conf["type"]
        repo_settings = {}
        for setting in repo_conf["settings"]:
            repo_settings[setting] = repo_conf["settings"][setting]
        return repo_name, {"type" : repo_type,"settings" : repo_settings}

    def create_repository(self, envconf, build_doc, index_meta):
        aws_key=envconf.get("cloud",{}).get("access_key")
        aws_secret=envconf.get("cloud",{}).get("secret_key")
        repo_name, repo_conf = self.get_repository_config(envconf["repository"],build_doc,index_meta)
        self.logger.info("Repository config: %s" % repo_conf)
        settings = repo_conf["settings"]
        repo_type = repo_conf["type"]
        es_idxr = self.get_es_idxr(envconf)
        try:
            es_idxr.get_repository(repo_name)
        except ESIndexerException:
            # need to create that repo
            if repo_type == "s3":
                acl = repo_conf.get("acl",None) # let aws.create_bucket default it
                # first make sure bucket exists
                aws.create_bucket(
                        name=settings["bucket"],
                        region=settings["region"],
                        aws_key=aws_key,aws_secret=aws_secret,
                        acl=acl,
                        ignore_already_exists=True)
            settings = {"type" : repo_type,
                        "settings" : settings}
            self.logger.info("Create repository named '%s': %s" % (repo_name,pformat(settings)))
            es_idxr.create_repository(repo_name,settings=settings)

        return repo_name,repo_conf


class SnapshotManager(BaseManager):

    DEFAULT_SNAPSHOOTER_CLASS = Snapshooter

    def __init__(self, index_manager, poll_schedule=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index_manager = index_manager
        self.t0 = time.time()
        self.log_folder = btconfig.LOG_FOLDER
        self.timestamp = datetime.now()
        self.snapshot_config = {}
        self.poll_schedule = poll_schedule
        self.es_backups_folder = btconfig.ES_BACKUPS_FOLDER
        self.setup()

    def clean_stale_status(self):
        src_build = get_src_build()
        for build in src_build.find():
            for job in build.get("jobs",[]):
                if job.get("status","").endswith("snapshotting"):
                    logging.warning("Found stale build '%s', marking snapshot status as 'canceled'" % build["_id"])
                    job["status"] = "canceled"
            src_build.replace_one({"_id":build["_id"]},build)

    def setup(self):
        self.setup_log()
    
    def setup_log(self):
        self.logger, self.logfile = get_logger(SNAPSHOTMANAGER_CATEGORY,self.log_folder)
    
    def poll(self,state,func):
        super().poll(state,func,col=get_src_build())
    
    def __getitem__(self,env):
        """
        Return an instance of a snapshooter for snapshot environment named "env"
        """
        d = self.register[env]
        envconf = self.snapshot_config["env"][env]
        return d["class"](index_manager=d["index_manager"],
                          job_manager=d["job_manager"],
                          envconf=envconf,log_folder=self.log_folder)

    def configure(self, snapshot_confdict):
        """
        Configure manager with snapshot config dict. See SNAPSHOT_CONFIG in config_hub.py
        for the format.
        """
        self.snapshot_config = copy.deepcopy(snapshot_confdict)
        for env, envconf in self.snapshot_config.get("env",{}).items():
            try:
                if envconf.get("cloud"):
                    assert envconf["cloud"]["type"] == "aws", \
                            "Only Amazon AWS cloud is supported at the moment"
                self.register[env] = {
                        "class" : self.DEFAULT_SNAPSHOOTER_CLASS,
                        "conf" : envconf,
                        "job_manager" : self.job_manager,
                        "index_manager" : self.index_manager
                        }
            except Exception as e:
                self.logger.exception("Couldn't setup snapshot environment '%s' because: %s" % (env,e))

    def snapshot(self, snapshot_env, index, snapshot=None, steps=["pre","snapshot","post"]):
        """
        Create a snapshot named "snapshot" (or, by default, same name as the index) from "index"
        according to environment definition (repository, etc...) "snapshot_env".

        """
        if not snapshot_env in self.snapshot_config.get("env",{}):
            raise ValueError("Unknonw snapshot environment '%s'" % snapshot_env)
        snapshooter = self[snapshot_env]
        return snapshooter.snapshot(index, snapshot=snapshot, steps=steps)

    def snapshot_info(self, env=None, remote=False):
        res = copy.deepcopy(self.snapshot_config)
        for kenv in self.snapshot_config["env"]:
            if env and env != kenv:
                continue
            if remote:
                raise NotImplementedError()
                # TODO: could list snapshot from ES
                #try:
                #    cl = Elasticsearch(res["env"][kenv]["host"],timeout=1,max_retries=0)
                #    indices = [{"index":k,
                #        "doc_type":list(v["mappings"].keys())[0],
                #        "aliases":list(v["aliases"].keys())}
                #        for k,v in cl.indices.get("*").items()]
                #    # for now, we just consider
                #    if type(res["env"][kenv]["index"]) == dict:
                #        # we don't where to put those indices because we don't
                #        # have that information, so we just put those in a default category
                #        # TODO: put that info in metadata ?
                #        res["env"][kenv]["index"].setdefault(None,[]).extend(indices)
                #    else:
                #        assert type(res["env"][kenv]["index"]) == list
                #        res["env"][kenv]["index"].extend(indices)
                #except Exception as e:
                #    self.logger.warning("Can't load remote indices: %s" % e)
                #    continue
        return res



from pprint import pprint, pformat
import asyncio, asyncssh, crypt, aiocron, os, sys, types, time, copy
from collections import OrderedDict
from functools import partial
import logging

from biothings import config
from biothings.utils.loggers import get_logger, WSLogHandler, WSShellHandler, ShellLogger
from biothings.utils.hub import HubShell, HubReloader, CommandDefinition, pending, \
                                AlreadyRunningException, CommandError
from biothings.utils.jsondiff import make as jsondiff

# Keys used as category in pinfo (description of jobs submitted to JobManager)
# Those are used in different places
DUMPER_CATEGORY = "dumper"
UPLOADER_CATEGORY = "uploader"
BUILDER_CATEGORY = "builder"
INDEXER_CATEGORY = "indexer"
INDEXMANAGER_CATEGORY = "indexmanager"
RELEASEMANAGER_CATEGORY = "releasemanager"
RELEASER_CATEGORY = "releaser"
SNAPSHOTMANAGER_CATEGORY = "snapshotmanager"
SNAPSHOOTER_CATEGORY = "snapshooter"
DIFFER_CATEGORY = "differ"
DIFFMANAGER_CATEGORY = "diffmanager"
SYNCER_CATEGORY = "syncer"
INSPECTOR_CATEGORY = "inspector"

HUB_REFRESH_COMMANDS = hasattr(config,"HUB_REFRESH_COMMANDS") and config.HUB_REFRESH_COMMANDS or "* * * * * *" # every sec


class JobRenderer(object):

    def __init__(self):
        self.rendered = {
                types.FunctionType : self.render_func,
                types.MethodType : self.render_method,
                partial : self.render_partial,
                types.LambdaType: self.render_lambda,
        }

    def render(self,job):
        r = self.rendered.get(type(job._callback))
        rstr = r(job._callback)
        delta = job._when - job._loop.time()
        days = None
        if delta > 86400:
            days = int(delta/86400)
            delta = delta - 86400
        strdelta = time.strftime("%Hh:%Mm:%Ss", time.gmtime(int(delta)))
        if days:
            strdelta = "%d day(s) %s" % (days,strdelta)
        return "%s {run in %s}" % (rstr,strdelta)

    def render_partial(self,p):
        # class.method(args)
        return self.rendered[type(p.func)](p.func) + "%s" % str(p.args)

    def render_cron(self,c):
        # func type associated to cron can vary
        return self.rendered[type(c.func)](c.func) + " [%s]" % c.spec

    def render_func(self,f):
        return f.__name__

    def render_method(self,m):
        # what is self ? cron ?
        if type(m.__self__) == aiocron.Cron:
            return self.render_cron(m.__self__)
        else:
            return "%s.%s" % (m.__self__.__class__.__name__,
                              m.__name__)

    def render_lambda(self,l):
        return l.__name__

renderer = JobRenderer()


def status(managers):
    """
    Return a global hub status (number or sources, documents, etc...)
    according to available managers
    """
    total_srcs = None
    total_docs = None
    total_confs = None
    total_builds = None
    total_apis = None
    total_running_apis = None
    if managers.get("source_manager"):
        try:
            srcm = managers["source_manager"]
            srcs = srcm.get_sources()
            total_srcs = len(srcs)
            total_docs = sum([s["upload"]["sources"][subs].get("count",0) or 0 \
                            for s in srcs
                            for subs in s.get("upload",{}).get("sources",{}) \
                            if s.get("upload")])
        except Exception as e:
            logging.error("Can't get stats for sources: %s" % e)

    try:
        bm = managers["build_manager"]
        total_confs = len(bm.build_config_info())
    except Exception as e:
        logging.error("Can't get total number of build configurations: %s" % e)
    try:
        total_builds = len(bm.build_info())
    except Exception as e:
        logging.error("Can't get total number of builds: %s" % e)

    try:
        am = managers["api_manager"]
        apis = am.get_apis()
        total_apis = len(apis)
        total_running_apis = len([a for a in apis if a.get("status") == "running"])
    except Exception as e:
        logging.error("Can't get stats for APIs: %s" % e)

    return {
            "source" : {"total" : total_srcs,
                        "documents" : total_docs},
            "build" : {"total" : total_builds},
            "build_conf" : {"total" : total_confs},
            "api" : {"total" : total_apis,
                     "running" : total_running_apis},
            }


def schedule(loop):
    jobs = {}
    # try to render job in a human-readable way...
    out = []
    for sch in loop._scheduled:
        if type(sch) != asyncio.events.TimerHandle:
            continue
        if sch._cancelled:
            continue
        try:
            info = renderer.render(sch)
            out.append(info)
        except Exception as e:
            import traceback
            traceback.print_exc()
            out.append(sch)

    return "\n".join(out)


@asyncio.coroutine
def start_ssh_server(loop,name,passwords,keys=['bin/ssh_host_key'],shell=None,
                 host='',port=8022):
    for key in keys:
        assert os.path.exists(key),"Missing key '%s' (use: 'ssh-keygen -f %s' to generate it" % (key,key)
    HubSSHServer.PASSWORDS = passwords
    HubSSHServer.NAME = name
    HubSSHServer.SHELL = shell
    cron = aiocron.crontab(HUB_REFRESH_COMMANDS,func=shell.__class__.refresh_commands,
                           start=True, loop=loop)
    yield from asyncssh.create_server(HubSSHServer, host, port, loop=loop,
                                 server_host_keys=keys)


class HubCommands(OrderedDict):

    def __setitem__(self,k,v):
        if k in self:
            raise ValueError("Command '%s' already defined" % k)
        super().__setitem__(k,v)


class HubServer(object):

    DEFAULT_FEATURES = ["config","job","dump","upload","dataplugin","source",
                        "build","diff","index","snapshot","release","inspect","sync","api",
                        "terminal","reloader","dataupload","ws"]
    DEFAULT_MANAGERS_ARGS = {"upload" : {"poll_schedule" : "* * * * * */10"}}
    DEFAULT_RELOADER_CONFIG = {"folders": None, # will use default one
                               "managers" : ["source_manager","assistant_manager"],
                               "reload_func" : None} # will use default one
    DEFAULT_DATAUPLOAD_CONFIG = {"upload_root" : getattr(config,"DATA_UPLOAD_FOLDER",None)}
    DEFAULT_WEBSOCKET_CONFIG = {}
    DEFAULT_API_CONFIG = {}

    def __init__(self, source_list, features=None, name="BioThings Hub",
                 managers_custom_args={}, api_config=None, reloader_config=None,
                 dataupload_config=None, websocket_config=None):
        """
        Helper to setup and instantiate common managers usually used in a hub
        (eg. dumper manager, uploader manager, etc...)
        "source_list" is either:
            - a list of string corresponding to paths to datasources modules
            - a package containing sub-folders with datasources modules
        Specific managers can be retrieved adjusting "features" parameter, where
        each feature corresponds to one or more managers. Parameter defaults to
        all possible available. Managers are configured/init in the same order as the list,
        so if a manager (eg. job_manager) is required by all others, it must be the first
        in the list.
        "managers_custom_args" is an optional dict used to pass specific arguments while
        init managers:
            managers_custom_args={"upload" : {"poll_schedule" : "*/5 * * * *"}}
        will set poll schedule to check upload every 5min (instead of default 10s)
        "reloader_config", "dataupload_config" and "websocket_config" can be used to
        customize reloader, dataupload and websocket. If None, default config is used.
        If explicitely False, feature is deactivated.
        """
        self.name = name
        self.source_list = source_list
        self.logger, self.logfile = get_logger("hub")
        self._passed_features = features
        self._passed_managers_custom_args = managers_custom_args
        self.features = self.clean_features(features or self.DEFAULT_FEATURES)
        self.managers_custom_args = managers_custom_args
        self.reloader_config = reloader_config or self.DEFAULT_RELOADER_CONFIG
        self.dataupload_config = dataupload_config or self.DEFAULT_DATAUPLOAD_CONFIG
        self.websocket_config = websocket_config or self.DEFAULT_WEBSOCKET_CONFIG
        self.ws_listeners = [] # collect listeners that should be connected (push data through) to websocket
        self.api_config = api_config or self.DEFAULT_API_CONFIG
        # set during configure()
        self.managers = None
        self.api_endpoints = None
        self.shell = None
        self.commands = None
        self.extra_commands = None
        self.routes = []
        # flag "do we need to configure?"
        self.configured = False

    def clean_features(self, features):
        # we can't just use "set()" because we need to preserve order
        ordered = OrderedDict()
        for feat in features:
            if not feat in ordered:
                ordered[feat] = None
        return list(ordered.keys())

    def before_configure(self):
        """
        Hook triggered before configure(),
        used eg. to adjust features list
        """
        pass

    def configure(self):
        self.before_configure()
        self.remaining_features = copy.deepcopy(self.features) # keep track of what's been configured
        self.configure_ioloop()
        self.configure_managers()
        # setup the shell
        self.shell = HubShell(self.managers["job_manager"])
        self.shell.register_managers(self.managers)
        self.shell.server = self # propagate server instance in shell
                                 # so it's accessible from the console if needed
        self.configure_remaining_features()
        self.configure_commands()
        self.configure_extra_commands()
        self.shell.set_commands(self.commands,self.extra_commands)
        # set api
        if self.api_config != False:
            self.configure_api_endpoints() # after shell setup as it adds some default commands
                                           # we want to expose throught the api
            from biothings.hub.api import generate_api_routes
            self.routes.extend(generate_api_routes(self.shell, self.api_endpoints))

        # done
        self.configured = True

    def configure_ioloop(self):
        import tornado.platform.asyncio
        tornado.platform.asyncio.AsyncIOMainLoop().install()

    def before_start(self):
        pass

    def start(self):
        if not self.configured:
            self.configure()
        self.logger.info("Starting server '%s'" % self.name)
        # can't use asyncio.get_event_loop() if python < 3.5.3 as it would return
        # another instance of aio loop, take it from job_manager to make sure
        # we share the same one
        loop = self.managers["job_manager"].loop
        if self.routes:
            self.logger.info(self.routes)
            self.logger.info("Starting Hub API server")
            import tornado.web
            # register app into current event loop
            api = tornado.web.Application(self.routes)
            self.extra_commands["api"] = api
            from biothings.hub.api import start_api
            api_server = start_api(api,config.HUB_API_PORT,settings=getattr(config,"TORNADO_SETTINGS",{}))
        else:
            self.logger.info("No route defined, API server won't start")
        # at this point, everything is ready/set, last call for customizations
        self.before_start()
        self.ssh_server = start_ssh_server(loop,self.name,passwords=config.HUB_PASSWD,
                              port=config.HUB_SSH_PORT,shell=self.shell)
        try:
            loop.run_until_complete(self.ssh_server)
        except (OSError, asyncssh.Error) as exc:
            sys.exit('Error starting server: ' + str(exc))
        loop.run_forever()

    def mixargs(self, feat, params={}):
        args = {}
        for p in params:
            args[p] = self.managers_custom_args.get(feat,{}).pop(p,None) or params[p]
        # mix remaining
        args.update(self.managers_custom_args.get(feat,{}))
        return args

    def configure_job_manager(self):
        import asyncio
        loop = asyncio.get_event_loop()
        from biothings.utils.manager import JobManager
        args = self.mixargs("job",{"num_workers":config.HUB_MAX_WORKERS,"max_memory_usage":config.HUB_MAX_MEM_USAGE})
        job_manager = JobManager(loop,**args)
        self.managers["job_manager"] = job_manager

    def configure_dump_manager(self):
        from biothings.hub.dataload.dumper import DumperManager
        args = self.mixargs("dump")
        dmanager = DumperManager(job_manager=self.managers["job_manager"],**args)
        self.managers["dump_manager"] = dmanager

    def configure_upload_manager(self):
        from biothings.hub.dataload.uploader import UploaderManager
        args = self.mixargs("upload",{"poll_schedule":"* * * * * */10"})
        upload_manager = UploaderManager(job_manager=self.managers["job_manager"],**args)
        self.managers["upload_manager"] = upload_manager

    def configure_dataplugin_manager(self):
        from biothings.hub.dataplugin.manager import DataPluginManager
        dp_manager = DataPluginManager(job_manager=self.managers["job_manager"])
        self.managers["dataplugin_manager"] = dp_manager
        from biothings.hub.dataplugin.assistant import AssistantManager
        args = self.mixargs("dataplugin")
        assistant_manager = AssistantManager(
                data_plugin_manager=dp_manager,
                dumper_manager=self.managers["dump_manager"],
                uploader_manager=self.managers["upload_manager"],
                job_manager=self.managers["job_manager"],
                **args)
        self.managers["assistant_manager"] = assistant_manager

    def configure_build_manager(self):
        from biothings.hub.databuild.builder import BuilderManager
        args = self.mixargs("build")
        build_manager = BuilderManager(job_manager=self.managers["job_manager"],**args)
        build_manager.configure()
        self.managers["build_manager"] = build_manager

    def configure_diff_manager(self):
        from biothings.hub.databuild.differ import DifferManager, SelfContainedJsonDiffer
        args = self.mixargs("diff")
        diff_manager = DifferManager(job_manager=self.managers["job_manager"],
                                     poll_schedule="* * * * * */10",**args)
        diff_manager.configure([SelfContainedJsonDiffer,])
        diff_manager.poll("diff",lambda doc: diff_manager.diff("jsondiff-selfcontained",old=None,new=doc["_id"]))
        self.managers["diff_manager"] = diff_manager

    def configure_index_manager(self):
        from biothings.hub.dataindex.indexer import IndexManager
        args = self.mixargs("index")
        index_manager = IndexManager(job_manager=self.managers["job_manager"],**args)
        index_manager.configure(config.INDEX_CONFIG)
        self.managers["index_manager"] = index_manager

    def configure_snapshot_manager(self):
        assert "index" in self.features, "'snapshot' feature requires 'index'"
        from biothings.hub.dataindex.snapshooter import SnapshotManager
        args = self.mixargs("snapshot")
        snapshot_manager = SnapshotManager(
                index_manager=self.managers["index_manager"],
                job_manager=self.managers["job_manager"],
                **args)
        snapshot_manager.configure(config.SNAPSHOT_CONFIG)
        #snapshot_manager.poll("snapshot",lambda doc: snapshot_manager.snapshot(snapshot_env=???,index=doc["_id"]))
        self.managers["snapshot_manager"] = snapshot_manager

    def configure_release_manager(self):
        assert "diff" in self.features, "'release' feature requires 'diff'"
        assert "snapshot" in self.features, "'release' feature requires 'snapshot'"
        from biothings.hub.datarelease.publisher import ReleaseManager
        args = self.mixargs("release")
        release_manager = ReleaseManager(
                diff_manager=self.managers["diff_manager"],
                snapshot_manager=self.managers["snapshot_manager"],
                job_manager=self.managers["job_manager"],
                poll_schedule="* * * * * */10",**args)
        release_manager.configure(config.RELEASE_CONFIG)
        release_manager.poll("release_note",lambda doc: release_manager.create_release_note(old=None,new=doc["_id"]))
        self.managers["release_manager"] = release_manager

    def configure_sync_manager(self):
        from biothings.hub.databuild.syncer import SyncerManager
        args = self.mixargs("sync")
        sync_manager = SyncerManager(job_manager=self.managers["job_manager"],**args)
        sync_manager.configure()
        self.managers["sync_manager"] = sync_manager

    def configure_inspect_manager(self):
        assert "upload" in self.features, "'inspect' feature requires 'upload'"
        assert "build" in self.features, "'inspect' feature requires 'build'"
        from biothings.hub.datainspect.inspector import InspectorManager
        args = self.mixargs("inspect")
        inspect_manager = InspectorManager(
                upload_manager=self.managers["upload_manager"],
                build_manager=self.managers["build_manager"],
                job_manager=self.managers["job_manager"],**args)
        self.managers["inspect_manager"] = inspect_manager

    def configure_api_manager(self):
        assert "index" in self.features, "'api' feature requires 'index'"
        from biothings.hub.api.manager import APIManager
        args = self.mixargs("api")
        api_manager = APIManager(**args)
        self.managers["api_manager"] = api_manager

    def configure_source_manager(self):
        if "dump" in self.features or "upload" in self.features:
            args = self.mixargs("source")
            from biothings.hub.dataload.source import SourceManager
            source_manager = SourceManager(
                    source_list=self.source_list,
                    dump_manager=self.managers["dump_manager"],
                    upload_manager=self.managers["upload_manager"],
                    data_plugin_manager=self.managers.get("dataplugin_manager"),
                    )
            self.managers["source_manager"] = source_manager
        # init data plugin once source_manager has been set (it inits dumper and uploader
        # managers, if assistant_manager is configured/loaded before, datasources won't appear
        # in dumper/uploader managers as they were not ready yet)
        if "dataplugin" in self.features:
            self.managers["assistant_manager"].configure()
            self.managers["assistant_manager"].load()

        # now that we have the source manager setup, we can schedule and poll
        if "dump" in self.features and not getattr(config,"SKIP_DUMPER_SCHEDULE",False):
            self.managers["dump_manager"].schedule_all()
        if "upload" in self.features and not getattr(config,"SKIP_UPLOADER_POLL",False):
            self.managers["upload_manager"].poll('upload',lambda doc:
                    self.shell.launch(partial(self.managers["upload_manager"].upload_src,doc["_id"])))

    def configure_managers(self):
        if not self.managers is None:
            raise Exception("Managers have already been configured")
        self.managers = {}

        self.logger.info("Setting up managers for following features: %s" % self.features)
        assert "job" in self.features, "'job' feature is mandatory"
        if "source" in self.features:
            assert "dump" in self.features and "upload" in self.features, \
                    "'source' feature requires both 'dump' and 'upload' features"
        if "dataplugin" in self.features:
            assert "source" in self.features, "'dataplugin' feature requires 'source' feature"

        # specific order, eg. job_manager is used by all managers
        for feat in self.features:
            if hasattr(self,"configure_%s_manager" % feat):
                getattr(self,"configure_%s_manager" % feat)()
                self.remaining_features.remove(feat)
            elif hasattr(self,"configure_%s_feature" % feat):
                # see configure_remaining_features()
                pass # this is configured after managers but should not produce an error
            else:
                raise AttributeError("Feature '%s' listed but no 'configure_%s_{manager|feature}' method found" % (feat,feat))

        self.logger.info("Active manager(s): %s" % pformat(self.managers))

    def configure_config_feature(self):
        # just a placeholder
        pass

    def configure_ws_feature(self):
        # add websocket endpoint
        import biothings.hub.api.handlers.ws as ws
        import sockjs.tornado
        from biothings.utils.hub_db import ChangeWatcher
        # monitor change in database to report activity in webapp
        self.db_listener = ws.HubDBListener()
        ChangeWatcher.add(self.db_listener)
        ChangeWatcher.publish()
        self.log_listener = ws.LogListener()
        # push log statements to the webapp
        root_logger = logging.getLogger() # careful, asyncio logger will trigger log statement while in the handler
                                          # (ie. infinite loop), root logger not recommended) 
        root_logger.addHandler(WSLogHandler(self.log_listener))
        self.ws_listeners.extend([self.db_listener,self.log_listener])

        ws_router = sockjs.tornado.SockJSRouter(
                partial(ws.WebSocketConnection,
                    listeners=self.ws_listeners),
                '/ws')
        self.routes.extend(ws_router.urls)

    def configure_terminal_feature(self):
        assert "ws" in self.features, "'terminal' feature requires 'ws'"
        assert "ws" in self.remaining_features, "'terminal' feature should configured before 'ws'"
        # shell logger/listener to communicate between webapp and hub ssh console
        import biothings.hub.api.handlers.ws as ws
        shell_listener = ws.LogListener()
        shell_logger = logging.getLogger("shell")
        assert isinstance(shell_logger,ShellLogger), "shell_logger isn't properly set"
        shell_logger.addHandler(WSShellHandler(shell_listener))
        self.ws_listeners.append(shell_listener)
        # webapp terminal to hub shell connection through /shell endpoint
        from biothings.hub.api.handlers.shell import ShellHandler
        shell_endpoint = ("/shell",ShellHandler,
                {"shell":self.shell,"shellog":shell_logger})
        self.routes.append(shell_endpoint)

    def configure_dataupload_feature(self):
        assert "ws" in self.features, "'dataupload' feature requires 'ws'"
        assert "ws" in self.remaining_features, "'dataupload' feature should configured before 'ws'"
        # this one is not bound to a specific command
        from biothings.hub.api.handlers.upload import UploadHandler
        # tuple type = interpreted as a route handler
        self.routes.append(("/dataupload/([\w\.-]+)?",UploadHandler,self.dataupload_config))

    def configure_reloader_feature(self):
        monitored_folders = self.reloader_config["folders"] or ["hub/dataload/sources",getattr(config,"DATA_PLUGIN_FOLDER",None)]
        reload_managers = [self.managers[m] for m in self.reloader_config["managers"] if m in self.managers]
        reload_func = self.reloader_config["reload_func"] or partial(self.shell.restart,force=True)
        reloader = HubReloader(monitored_folders, reload_managers, reload_func=reload_func)
        reloader.monitor()

    def configure_remaining_features(self):
        self.logger.info("Setting up remaining features: %s" % self.remaining_features)
        # specific order, eg. job_manager is used by all managers
        for feat in copy.deepcopy(self.remaining_features):
            if hasattr(self,"configure_%s_feature" % feat):
                getattr(self,"configure_%s_feature" % feat)()
                self.remaining_features.remove(feat)
                pass # this is configured after managers but should not produce an error
            else:
                raise AttributeError("Feature '%s' listed but no 'configure_%s_feature' method found" % (feat,feat))
        
    def configure_commands(self):
        """
        Configure hub commands according to available managers
        """
        assert self.managers, "No managers configured"
        self.commands = HubCommands()
        self.commands["status"] = CommandDefinition(command=partial(status,self.managers),tracked=False)
        if "config" in self.features:
            self.commands["config"] = CommandDefinition(command=config.show,tracked=False)
            self.commands["setconf"] = config.store_value_to_db
            self.commands["resetconf"] = config.reset
        # getting info
        if self.managers.get("source_manager"):
            self.commands["source_info"] = CommandDefinition(command=self.managers["source_manager"].get_source,tracked=False)
        # dump commands
        if self.managers.get("dump_manager"):
            self.commands["dump"] = self.managers["dump_manager"].dump_src
            self.commands["dump_all"] = self.managers["dump_manager"].dump_all
        # upload commands
        if self.managers.get("upload_manager"):
            self.commands["upload"] = self.managers["upload_manager"].upload_src
            self.commands["upload_all"] = self.managers["upload_manager"].upload_all
        # building/merging
        if self.managers.get("build_manager"):
            self.commands["whatsnew"] = CommandDefinition(command=self.managers["build_manager"].whatsnew,tracked=False)
            self.commands["lsmerge"] = self.managers["build_manager"].list_merge
            self.commands["rmmerge"] = self.managers["build_manager"].delete_merge
            self.commands["merge"] = self.managers["build_manager"].merge
            self.commands["archive"] = self.managers["build_manager"].archive_merge
        if hasattr(config,"INDEX_CONFIG"):
            self.commands["index_config"] = config.INDEX_CONFIG
        if hasattr(config,"SNAPSHOT_CONFIG"):
            self.commands["snapshot_config"] = config.SNAPSHOT_CONFIG
        if hasattr(config,"PUBLISH_CONFIG"):
            self.commands["publish_config"] = config.PUBLISH_CONFIG
        # diff
        if self.managers.get("diff_manager"):
            self.commands["diff"] = self.managers["diff_manager"].diff
            self.commands["report"] = self.managers["diff_manager"].diff_report
        # indexing commands
        if self.managers.get("index_manager"):
            self.commands["index"] = self.managers["index_manager"].index
        if self.managers.get("snapshot_manager"):
            self.commands["snapshot"] = self.managers["snapshot_manager"].snapshot
        # data release commands
        if self.managers.get("release_manager"):
            self.commands["create_release_note"] = self.managers["release_manager"].create_release_note
            self.commands["get_release_note"] = CommandDefinition(command=self.managers["release_manager"].get_release_note,tracked=False)
            self.commands["publish"] = self.managers["release_manager"].publish
            self.commands["publish_diff"] = self.managers["release_manager"].publish_diff
            self.commands["publish_snapshot"] = self.managers["release_manager"].publish_snapshot
        if self.managers.get("sync_manager"):
            self.commands["sync"] = CommandDefinition(command=self.managers["sync_manager"].sync)
        # inspector
        if self.managers.get("inspect_manager"):
            self.commands["inspect"] = self.managers["inspect_manager"].inspect
        # data plugins
        if self.managers.get("assistant_manager"):
            self.commands["register_url"] = partial(self.managers["assistant_manager"].register_url)
            self.commands["unregister_url"] = partial(self.managers["assistant_manager"].unregister_url)
            self.commands["export_plugin"] = partial(self.managers["assistant_manager"].export)
        if self.managers.get("dataplugin_manager"):
            self.commands["dump_plugin"] = self.managers["dataplugin_manager"].dump_src

        logging.info("Registered commands: %s" % list(self.commands.keys()))

    def configure_extra_commands(self):
        """
        Same as configure_commands() but commands are not exposed publicly in the shell
        (they are shortcuts or commands for API endpoints, supporting commands, etc...)
        """
        assert self.managers, "No managers configured"
        self.extra_commands = {} # unordered since not exposed, we don't care
        loop = self.managers.get("job_manager") and self.managers["job_manager"].loop or asyncio.get_event_loop()
        self.extra_commands["g"] = CommandDefinition(command=globals(),tracked=False)
        self.extra_commands["sch"] = CommandDefinition(command=partial(schedule,loop),tracked=False)
        # expose contant so no need to put quotes (eg. top(pending) instead of top("pending")
        self.extra_commands["pending"] = CommandDefinition(command=pending,tracked=False)
        self.extra_commands["loop"] = CommandDefinition(command=loop,tracked=False)

        if self.managers.get("job_manager"):
            self.extra_commands["pqueue"] = CommandDefinition(command=self.managers["job_manager"].process_queue,tracked=False)
            self.extra_commands["tqueue"] = CommandDefinition(command=self.managers["job_manager"].thread_queue,tracked=False)
            self.extra_commands["jm"] = CommandDefinition(command=self.managers["job_manager"],tracked=False)
            self.extra_commands["top"] = CommandDefinition(command=self.managers["job_manager"].top,tracked=False)
            self.extra_commands["job_info"] = CommandDefinition(command=self.managers["job_manager"].job_info,tracked=False)
        if self.managers.get("source_manager"):
            self.extra_commands["sm"] = CommandDefinition(command=self.managers["source_manager"],tracked=False)
            self.extra_commands["sources"] = CommandDefinition(command=self.managers["source_manager"].get_sources,tracked=False)
            self.extra_commands["source_save_mapping"] = CommandDefinition(command=self.managers["source_manager"].save_mapping)
        if self.managers.get("dump_manager"):
            self.extra_commands["dm"] = CommandDefinition(command=self.managers["dump_manager"],tracked=False)
            self.extra_commands["dump_info"] = CommandDefinition(command=self.managers["dump_manager"].dump_info,tracked=False)
        if self.managers.get("dataplugin_manager"):
            self.extra_commands["dpm"] = CommandDefinition(command=self.managers["dataplugin_manager"],tracked=False)
        if self.managers.get("assistant_manager"):
            self.extra_commands["am"] = CommandDefinition(command=self.managers["assistant_manager"],tracked=False)
        if self.managers.get("upload_manager"):
            self.extra_commands["um"] = CommandDefinition(command=self.managers["upload_manager"],tracked=False)
            self.extra_commands["upload_info"] = CommandDefinition(command=self.managers["upload_manager"].upload_info,tracked=False)
        if self.managers.get("build_manager"):
            self.extra_commands["bm"] = CommandDefinition(command=self.managers["build_manager"],tracked=False)
            self.extra_commands["builds"] = CommandDefinition(command=self.managers["build_manager"].build_info,tracked=False)
            self.extra_commands["build"] = CommandDefinition(command=lambda id: self.managers["build_manager"].build_info(id=id),tracked=False)
            self.extra_commands["build_config_info"] = CommandDefinition(command=self.managers["build_manager"].build_config_info,tracked=False)
            self.extra_commands["build_save_mapping"] = CommandDefinition(command=self.managers["build_manager"].save_mapping)
            self.extra_commands["create_build_conf"] = CommandDefinition(command=self.managers["build_manager"].create_build_configuration)
            self.extra_commands["update_build_conf"] = CommandDefinition(command=self.managers["build_manager"].update_build_configuration)
            self.extra_commands["delete_build_conf"] = CommandDefinition(command=self.managers["build_manager"].delete_build_configuration)
        if self.managers.get("diff_manager"):
            self.extra_commands["dim"] = CommandDefinition(command=self.managers["diff_manager"],tracked=False)
            self.extra_commands["diff_info"] = CommandDefinition(command=self.managers["diff_manager"].diff_info,tracked=False)
            self.extra_commands["jsondiff"] = CommandDefinition(command=jsondiff,tracked=False)
        if self.managers.get("sync_manager"):
            self.extra_commands["sym"] = CommandDefinition(command=self.managers["sync_manager"],tracked=False)
        if self.managers.get("index_manager"):
            self.extra_commands["im"] = CommandDefinition(command=self.managers["index_manager"],tracked=False)
            self.extra_commands["index_info"] = CommandDefinition(command=self.managers["index_manager"].index_info,tracked=False)
            self.extra_commands["validate_mapping"] = CommandDefinition(command=self.managers["index_manager"].validate_mapping)
        if self.managers.get("snapshot_manager"):
            self.extra_commands["ssm"] = CommandDefinition(command=self.managers["snapshot_manager"],tracked=False)
            self.extra_commands["snapshot_info"] = CommandDefinition(command=self.managers["snapshot_manager"].snapshot_info,tracked=False)
        if self.managers.get("release_manager"):
            self.extra_commands["rm"] = CommandDefinition(command=self.managers["release_manager"],tracked=False)
            self.extra_commands["release_info"] = CommandDefinition(command=self.managers["release_manager"].release_info,tracked=False)
        if self.managers.get("inspect_manager"):
            self.extra_commands["ism"] = CommandDefinition(command=self.managers["inspect_manager"],tracked=False)
        if self.managers.get("api_manager"):
            self.extra_commands["api"] = CommandDefinition(command=self.managers["api_manager"],tracked=False)
            self.extra_commands["get_apis"] = CommandDefinition(command=self.managers["api_manager"].get_apis,tracked=False)
            self.extra_commands["delete_api"] = CommandDefinition(command=self.managers["api_manager"].delete_api)
            self.extra_commands["create_api"] = CommandDefinition(command=self.managers["api_manager"].create_api)
            self.extra_commands["start_api"] = CommandDefinition(command=self.managers["api_manager"].start_api)
            self.extra_commands["stop_api"] = self.managers["api_manager"].stop_api

        logging.debug("Registered extra (private) commands: %s" % list(self.extra_commands.keys()))

    def configure_api_endpoints(self):
        cmdnames = list(self.commands.keys())
        if self.extra_commands:
            cmdnames.extend(list(self.extra_commands.keys()))
        from biothings.hub.api import EndpointDefinition
        self.api_endpoints = {}
        self.api_endpoints["config"] = []
        if "config" in cmdnames:
            self.api_endpoints["config"].append(EndpointDefinition(name="config",method="get"))
            self.api_endpoints["config"].append(EndpointDefinition(name="setconf",method="put",force_bodyargs=True))
            self.api_endpoints["config"].append(EndpointDefinition(name="resetconf",method="delete",force_bodyargs=True))
        if not self.api_endpoints["config"]:
            self.api_endpoints.pop("config")
        if "builds" in cmdnames: self.api_endpoints["builds"] = EndpointDefinition(name="builds",method="get")
        self.api_endpoints["build"] = []
        if "build" in cmdnames: self.api_endpoints["build"].append(EndpointDefinition(method="get",name="build"))
        if "archive" in cmdnames: self.api_endpoints["build"].append(EndpointDefinition(method="post",name="archive",suffix="archive"))
        if "rmmerge" in cmdnames: self.api_endpoints["build"].append(EndpointDefinition(method="delete",name="rmmerge"))
        if "merge" in cmdnames: self.api_endpoints["build"].append(EndpointDefinition(name="merge",method="put",suffix="new"))
        if "build_save_mapping" in cmdnames: self.api_endpoints["build"].append(EndpointDefinition(name="build_save_mapping",method="put",suffix="mapping"))
        if not self.api_endpoints["build"]:
            self.api_endpoints.pop("build")
        self.api_endpoints["publish"] = []
        if "publish_diff" in cmdnames: self.api_endpoints["publish"].append(EndpointDefinition(name="publish_diff",method="post",suffix="incremental",force_bodyargs=True))
        if "publish_snapshot" in cmdnames: self.api_endpoints["publish"].append(EndpointDefinition(name="publish_snapshot",method="post",suffix="full",force_bodyargs=True))
        if not self.api_endpoints["publish"]:
            self.api_endpoints.pop("publish")
        if "diff" in cmdnames: self.api_endpoints["diff"] = EndpointDefinition(name="diff",method="put",force_bodyargs=True)
        if "job_info" in cmdnames: self.api_endpoints["job_manager"] = EndpointDefinition(name="job_info",method="get")
        if "dump_info" in cmdnames: self.api_endpoints["dump_manager"] = EndpointDefinition(name="dump_info", method="get")
        if "upload_info" in cmdnames: self.api_endpoints["upload_manager"] = EndpointDefinition(name="upload_info",method="get")
        if "build_config_info" in cmdnames: self.api_endpoints["build_manager"] = EndpointDefinition(name="build_config_info",method="get")
        if "index_info" in cmdnames: self.api_endpoints["index_manager"] = EndpointDefinition(name="index_info",method="get")
        if "snapshot_info" in cmdnames: self.api_endpoints["snapshot_manager"] = EndpointDefinition(name="snapshot_info",method="get")
        if "release_info" in cmdnames: self.api_endpoints["release_manager"] = EndpointDefinition(name="release_info",method="get")
        if "diff_info" in cmdnames: self.api_endpoints["diff_manager"] = EndpointDefinition(name="diff_info",method="get")
        if "commands" in cmdnames: self.api_endpoints["commands"] = EndpointDefinition(name="commands",method="get")
        if "command" in cmdnames: self.api_endpoints["command"] = EndpointDefinition(name="command",method="get")
        if "sources" in cmdnames: self.api_endpoints["sources"] = EndpointDefinition(name="sources",method="get")
        self.api_endpoints["source"] = []
        if "source_info" in cmdnames: self.api_endpoints["source"].append(EndpointDefinition(name="source_info",method="get"))
        if "dump" in cmdnames: self.api_endpoints["source"].append(EndpointDefinition(name="dump",method="put",suffix="dump"))
        if "upload" in cmdnames: self.api_endpoints["source"].append(EndpointDefinition(name="upload",method="put",suffix="upload"))
        if "source_save_mapping" in cmdnames: self.api_endpoints["source"].append(EndpointDefinition(name="source_save_mapping",method="put",suffix="mapping"))
        if not self.api_endpoints["source"]:
            self.api_endpoints.pop("source")
        if "inspect" in cmdnames: self.api_endpoints["inspect"] = EndpointDefinition(name="inspect",method="put",force_bodyargs=True)
        if "register_url" in cmdnames: self.api_endpoints["dataplugin/register_url"] = EndpointDefinition(name="register_url",method="post",force_bodyargs=True)
        if "unregister_url" in cmdnames: self.api_endpoints["dataplugin/unregister_url"] = EndpointDefinition(name="unregister_url",method="delete",force_bodyargs=True)
        self.api_endpoints["dataplugin"] = []
        if "dump_plugin" in cmdnames: self.api_endpoints["dataplugin"].append(EndpointDefinition(name="dump_plugin",method="put",suffix="dump"))
        if "export_plugin" in cmdnames: self.api_endpoints["dataplugin"].append(EndpointDefinition(name="export_plugin",method="put",suffix="export"))
        if not self.api_endpoints["dataplugin"]:
            self.api_endpoints.pop("dataplugin")
        if "jsondiff" in cmdnames: self.api_endpoints["jsondiff"] = EndpointDefinition(name="jsondiff",method="post",force_bodyargs=True)
        if "validate_mapping" in cmdnames: self.api_endpoints["mapping/validate"] = EndpointDefinition(name="validate_mapping",method="post",force_bodyargs=True)
        self.api_endpoints["buildconf"] = []
        if "create_build_conf" in cmdnames:
            self.api_endpoints["buildconf"].append(EndpointDefinition(name="create_build_conf",method="post",force_bodyargs=True))
            self.api_endpoints["buildconf"].append(EndpointDefinition(name="update_build_conf",method="put",force_bodyargs=True))
        if "delete_build_conf" in cmdnames: self.api_endpoints["buildconf"].append(EndpointDefinition(name="delete_build_conf",method="delete",force_bodyargs=True))
        if not self.api_endpoints["buildconf"]:
            self.api_endpoints.pop("buildconf")
        if "index" in cmdnames: self.api_endpoints["index"] = EndpointDefinition(name="index",method="put",force_bodyargs=True)
        if "snapshot" in cmdnames: self.api_endpoints["snapshot"] = EndpointDefinition(name="snapshot",method="put",force_bodyargs=True)
        if "sync" in cmdnames: self.api_endpoints["sync"] = EndpointDefinition(name="sync",method="post",force_bodyargs=True)
        if "whatsnew" in cmdnames: self.api_endpoints["whatsnew"] = EndpointDefinition(name="whatsnew",method="get")
        if "status" in cmdnames: self.api_endpoints["status"] = EndpointDefinition(name="status",method="get")
        self.api_endpoints["release_note"] = []
        if "create_release_note" in cmdnames:
            self.api_endpoints["release_note"].append(EndpointDefinition(name="create_release_note",method="put",suffix="create",force_bodyargs=True))
        if "get_release_note" in cmdnames:
            self.api_endpoints["release_note"].append(EndpointDefinition(name="get_release_note",method="get",force_bodyargs=True))
        if not self.api_endpoints["release_note"]:
            self.api_endpoints.pop("release_note")
        self.api_endpoints["api"] = []
        if "start_api" in cmdnames: self.api_endpoints["api"].append(EndpointDefinition(name="start_api",method="put",suffix="start"))
        if "stop_api" in cmdnames: self.api_endpoints["api"].append(EndpointDefinition(name="stop_api",method="put",suffix="stop"))
        if "delete_api" in cmdnames: self.api_endpoints["api"].append(EndpointDefinition(name="delete_api",method="delete",force_bodyargs=True))
        if "create_api" in cmdnames: self.api_endpoints["api"].append(EndpointDefinition(name="create_api",method="post",force_bodyargs=True))
        if not self.api_endpoints["api"]:
            self.api_endpoints.pop("api")
        if "get_apis" in cmdnames: self.api_endpoints["api/list"] = EndpointDefinition(name="get_apis",method="get")
        if "stop" in cmdnames: self.api_endpoints["stop"] = EndpointDefinition(name="stop",method="put")
        if "restart" in cmdnames: self.api_endpoints["restart"] = EndpointDefinition(name="restart",method="put")


class HubSSHServer(asyncssh.SSHServer):

    PASSWORDS = {}
    SHELL = None

    def session_requested(self):
        return HubSSHServerSession(self.__class__.NAME,self.__class__.SHELL)

    def connection_made(self, conn):
         self._conn = conn
         print('SSH connection received from %s.' %
         conn.get_extra_info('peername')[0])

    def connection_lost(self, exc):
        if exc:
            print('SSH connection error: ' + str(exc), file=sys.stderr)
        else:
            print('SSH connection closed.')

    def begin_auth(self, username):
        try:
            self._conn.set_authorized_keys('bin/authorized_keys/%s.pub' % username)
        except IOError:
            pass
        return True

    def password_auth_supported(self):
        return True

    def validate_password(self, username, password):
        if self.password_auth_supported():
            pw = self.__class__.PASSWORDS.get(username, '*')
            return crypt.crypt(password, pw) == pw
        else:
            return False


class HubSSHServerSession(asyncssh.SSHServerSession):

    def __init__(self, name, shell):
        self.name = name
        self.shell = shell
        self._input = ''

    def connection_made(self, chan):
        self._chan = chan

    def shell_requested(self):
        return True

    def exec_requested(self,command):
        self.eval_lines(["%s" % command,"\n"])
        return True

    def session_started(self):
        welcome = ('\nWelcome to %s, %s!\n' % (self.name,self._chan.get_extra_info('username')))
        self.shell.shellog.output(welcome)
        self._chan.write(welcome)
        prompt = 'hub> '
        self.shell.shellog.output(prompt)
        self._chan.write(prompt)

    def data_received(self, data, datatype):
        self._input += data
        return self.eval_lines(self._input.split('\n'))

    def eval_lines(self, lines):
        for line in lines[:-1]:
            try:
                outs = [out for out in self.shell.eval(line) if out]
                # trailing \n if not already there
                if outs:
                    strout = "\n".join(outs).strip("\n") + "\n"
                    self._chan.write(strout)
                    self.shell.shellog.output(strout)
            except AlreadyRunningException as e:
                self._chan.write("AlreadyRunningException: %s" % e)
            except CommandError as e:
                self._chan.write("CommandError: %s" % e)
        self._chan.write('hub> ')
        # consume passed commands
        self._input = lines[-1]

    def eof_received(self):
        self._chan.write('Have a good one...\n')
        self._chan.exit(0)

    def break_received(self, msec):
        # simulate CR
        self._chan.write('\n')
        self.data_received("\n",None)




''' Biothings Test Utility '''

import os
import sys

from .web.helper import BiothingsTestCase, TornadoTestServerMixin

# Add Project Folder to PYTHONPATH
SRC_PATH = os.path.dirname(sys.path[0])
if SRC_PATH not in sys.path:
    sys.path.insert(1, SRC_PATH)



from tornado.web import HTTPError
from biothings.web.api.es.handlers.base_handler import BaseESRequestHandler
from biothings.web.api.es.query import BiothingSearchError
from biothings.utils.web import sum_arg_dicts
from biothings.web.api.helper import BiothingParameterTypeError
import logging
import traceback

class BiothingHandler(BaseESRequestHandler):
    ''' Request handlers for requests to the annotation lookup endpoint '''

    def initialize(self, web_settings):
        ''' Tornado handler `.initialize() <http://www.tornadoweb.org/en/stable/web.html#tornado.web.RequestHandler.initialize>`_ function for all requests to the annotation lookup endpoint.
        Here, the allowed arguments are set (depending on the request method) for each kwarg category.'''
        super(BiothingHandler, self).initialize(web_settings)
        self.ga_event_object_ret['action'] = self.request.method
        if self.request.method == 'GET':
            self.ga_event_object_ret['action'] = self.web_settings.GA_ACTION_ANNOTATION_GET
            self.control_kwargs = self.web_settings.ANNOTATION_GET_CONTROL_KWARGS
            self.es_kwargs = self.web_settings.ANNOTATION_GET_ES_KWARGS
            self.esqb_kwargs = self.web_settings.ANNOTATION_GET_ESQB_KWARGS
            self.transform_kwargs = self.web_settings.ANNOTATION_GET_TRANSFORM_KWARGS
        elif self.request.method == 'POST':
            self.ga_event_object_ret['action'] = self.web_settings.GA_ACTION_ANNOTATION_POST
            self.control_kwargs = self.web_settings.ANNOTATION_POST_CONTROL_KWARGS
            self.es_kwargs = self.web_settings.ANNOTATION_POST_ES_KWARGS
            self.esqb_kwargs = self.web_settings.ANNOTATION_POST_ESQB_KWARGS
            self.transform_kwargs = self.web_settings.ANNOTATION_POST_TRANSFORM_KWARGS
        else:
            # handle other verbs?
            pass
        self.kwarg_settings = sum_arg_dicts(self.control_kwargs, self.es_kwargs,
                                            self.esqb_kwargs, self.transform_kwargs)
        logging.debug("BiothingHandler - {}".format(self.request.method))
        logging.debug("Google Analytics Base object: {}".format(self.ga_event_object_ret))
        logging.debug("Kwarg settings: {}".format(self.kwarg_settings))

    def _regex_redirect(self, bid):
        ''' subclass to redirect based on a regex pattern (or whatever)... if this returns something falsy,
            the handler execution will continue past the redirect, otherwise if this returns something truthy
            the execution will return after the redirect (not sure about this) '''
        return False

    async def get(self, bid=None):
        ''' Handle a GET to the annotation lookup endpoint.'''
        if not bid:
            self.return_object(
                {'success': False, 'error': self.web_settings.ID_REQUIRED_MESSAGE},
                status_code=404)
            return

        # redirect this id
        if self._regex_redirect(bid):
            return

        ###################################################
        #              Get query parameters
        ###################################################

        # get kwargs from query and sanitize them
        try:
            kwargs = self.get_query_params()
        except BiothingParameterTypeError as e:
            self._return_data_and_track(
                {'success': False, 'error': "{0}".format(e)},
                ga_event_data={'qsize': 0},
                status_code=400)
            return

        ###################################################
        #           Split kwargs into categories
        ###################################################

        # split kwargs into options
        options = self.get_cleaned_options(kwargs)

        logging.debug("Request kwargs: {}".format(kwargs))
        logging.debug("Request options: {}".format(options))

        options = self._pre_query_builder_GET_hook(options)

        ###################################################
        #           Instantiate pipeline classes
        ###################################################

        # Instantiate query builder, query and transform classes
        _query_builder = self.web_settings.ES_QUERY_BUILDER(
            options=options.esqb_kwargs,
            regex_list=self.web_settings.ANNOTATION_ID_REGEX_LIST,
            index=self._get_es_index(options),
            doc_type=self._get_es_doc_type(options),
            es_options=options.es_kwargs,
            default_scopes=self.web_settings.DEFAULT_SCOPES)
        _backend = self.web_settings.ES_QUERY(
            client=self.web_settings.async_es_client,
            options=options.es_kwargs)
        _result_transformer = self.web_settings.ES_RESULT_TRANSFORMER(
            options=options.transform_kwargs,
            host=self.request.host,
            doc_url_function=self.web_settings.doc_url,
            output_aliases=self.web_settings.OUTPUT_KEY_ALIASES,
            jsonld_context=self.web_settings._jsonld_context,
            source_metadata=self.web_settings.source_metadata(),
            licenses=self.web_settings.LICENSE_TRANSFORM
        )

        ###################################################
        #                Build query
        ###################################################

        # get the query for annotation GET handler
        _query = _query_builder.annotation_GET_query(bid)

        logging.debug("Request query kwargs: {}".format(_query))

        # return raw query, if requested
        if options.control_kwargs.rawquery:
            self._return_data_and_track(
                _query.get('body', {'GET': bid}),
                rawquery=True, _format=options.control_kwargs.out_format)
            return

        _query = self._pre_query_GET_hook(options, _query)

        ###################################################
        #               Execute query
        ###################################################

        try:
            res = await _backend.annotation_GET_query(_query)
        except Exception:
            self.log_exceptions("Error executing query")
            self.return_object({'success': False, 'error': self.web_settings.ID_NOT_FOUND_TEMPLATE.format(
                bid=bid)}, status_code=404, _format=options.control_kwargs.out_format)
            #raise HTTPError(404)
            return

        #logging.debug("Raw query result: {}".format(res))

        # return raw result if requested
        if options.control_kwargs.raw:
            self._return_data_and_track(res, _format=options.control_kwargs.out_format)
            return

        res = self._pre_transform_GET_hook(options, res)

        ###################################################
        #           Transforming query result
        ###################################################

        # clean result
        try:
            res = _result_transformer.clean_annotation_GET_response(res)
        except Exception:
            self.log_exceptions("Error transforming result")
            self.return_object({'success': False, 'error': self.web_settings.ID_NOT_FOUND_TEMPLATE.format(
                bid=bid)}, status_code=404, _format=options.control_kwargs.out_format)
            #raise HTTPError(404)
            return

        # return result
        if not res:
            self.return_object({'success': False, 'error': self.web_settings.ID_NOT_FOUND_TEMPLATE.format(
                bid=bid)}, status_code=404, _format=options.control_kwargs.out_format)
            #raise HTTPError(404)
            return

        res = self._pre_finish_GET_hook(options, res)

        self._return_data_and_track(res, _format=options.control_kwargs.out_format)

    ###########################################################################

    async def post(self, ids=None):
        ''' Handle a POST to the annotation lookup endpoint '''

        ###################################################
        #           Get query parameters
        ###################################################

        try:
            kwargs = self.get_query_params()
        except BiothingParameterTypeError as e:
            self._return_data_and_track(
                {'success': False, 'error': "{0}".format(e)},
                ga_event_data={'qsize': 0},
                status_code=400)
            return
        #except Exception as e:
        #    self.log_exceptions("Error in get_query_params")
        #    self._return_data_and_track({'success': False, 'error': "Error parsing input parameter, check input types"}, ga_event_data={'qsize': 0})
        #    return

        # split kwargs into options
        options = self.get_cleaned_options(kwargs)

        logging.debug("Request kwargs: {}".format(kwargs))
        logging.debug("Request options: {}".format(options))

        if not options.control_kwargs.ids:
            self._return_data_and_track(
                {'success': False, 'error': "Missing required parameters."},
                ga_event_data={'qsize': 0},
                status_code=400, _format=options.control_kwargs.out_format)
            return

        options = self._pre_query_builder_POST_hook(options)

        ###################################################
        #           Instantiate pipeline classes
        ###################################################

        _query_builder = self.web_settings.ES_QUERY_BUILDER(
            options=options.esqb_kwargs,
            regex_list=self.web_settings.ANNOTATION_ID_REGEX_LIST,
            index=self._get_es_index(options),
            doc_type=self._get_es_doc_type(options),
            es_options=options.es_kwargs,
            default_scopes=self.web_settings.DEFAULT_SCOPES)
        _backend = self.web_settings.ES_QUERY(
            client=self.web_settings.async_es_client,
            options=options.es_kwargs)
        _result_transformer = self.web_settings.ES_RESULT_TRANSFORMER(
            options=options.transform_kwargs, host=self.request.host,
            doc_url_function=self.web_settings.doc_url,
            jsonld_context=self.web_settings._jsonld_context,
            output_aliases=self.web_settings.OUTPUT_KEY_ALIASES,
            source_metadata=self.web_settings.source_metadata(),
            licenses=self.web_settings.LICENSE_TRANSFORM
        )

        ###################################################
        #           Build query
        ###################################################

        _query = _query_builder.annotation_POST_query(options.control_kwargs.ids)
        #try:
        #    _query = _query_builder.annotation_POST_query(options.control_kwargs.ids)
        #except Exception as e:
        #    self.log_exceptions("Error building annotation POST query")
        #    self._return_data_and_track({'success': False, 'error': 'Error building query'}, ga_event_data={'qsize': len(options.control_kwargs.ids)})
        #    return

        logging.debug("Request query: {}".format(_query))

        if options.control_kwargs.rawquery:
            self._return_data_and_track(
                _query,
                ga_event_data={
                    'qsize': len(
                        options.control_kwargs.ids)},
                rawquery=True,
                _format=options.control_kwargs.out_format)
            return

        _query = self._pre_query_POST_hook(options, _query)

        ###################################################
        #           Execute query
        ###################################################

        try:
            res = await _backend.annotation_POST_query(_query)
        except TypeError as e:
            self.log_exceptions("Error executing annotation POST query")
            self._return_data_and_track(
                {'success': False, 'error': 'Error executing query'},
                ga_event_data={'qsize': len(options.control_kwargs.ids)},
                status_code=400, _format=options.control_kwargs.out_format)
            return
        except BiothingSearchError as e:
            self._return_data_and_track(
                {'success': False, 'error': '{0}'.format(e)},
                ga_event_data={'qsize': len(options.control_kwargs.ids)},
                status_code=400, _format=options.control_kwargs.out_format)
            return

        #logging.debug("Raw query result: {}".format(res))

        # return raw result if requested
        if options.control_kwargs.raw:
            self._return_data_and_track(
                res, ga_event_data={'qsize': len(options.control_kwargs.ids)},
                _format=options.control_kwargs.out_format)
            return

        res = self._pre_transform_POST_hook(options, res)

        ###################################################
        #           Transform query results
        ###################################################

        # clean result
        res = _result_transformer.clean_annotation_POST_response(
            bid_list=options.control_kwargs.ids, res=res)
        #try:
        #    res = _result_transformer.clean_annotation_POST_response(bid_list=options.control_kwargs.ids, res=res)
        #except Exception as e:
        #    self.log_exceptions("Error transforming annotation POST results")
        #    self._return_data_and_track({'success': False, 'error': 'Error transforming results'},
        #                    ga_event_data={'qsize': len(options.control_kwargs.ids)})

        res = self._pre_finish_POST_hook(options, res)

        # return and track
        self._return_data_and_track(
            res, ga_event_data={'qsize': len(options.control_kwargs.ids)},
            _format=options.control_kwargs.out_format)



from biothings.hub.datatransform.datatransform import DataTransform
from biothings.hub.datatransform.datatransform import IDStruct
from biothings.hub.datatransform.datatransform import DataTransformEdge
from biothings.hub.datatransform.datatransform import RegExEdge
from biothings.hub.datatransform.datatransform import nested_lookup

# only for testing purpose, this one is not optimized
#from biothings.hub.datatransform.datatransform_serial import DataTransformSerial

from biothings.hub.datatransform.datatransform_mdb import DataTransformMDB
from biothings.hub.datatransform.datatransform_mdb import MongoDBEdge
from biothings.hub.datatransform.datatransform_mdb import CIMongoDBEdge

# case-insensitive IDStruct class
from biothings.hub.datatransform.ciidstruct import CIIDStruct

# this involved biothings client dependency, skip it for now
#from biothings.hub.datatransform.datatransform_api import MyChemInfoEdge
#from biothings.hub.datatransform.datatransform_api import MyGeneInfoEdge



'''
    Biothings Query Component Common Tests
'''

import os

from nose.core import main

from biothings.tests import BiothingsTestCase


class QueryTests(BiothingsTestCase):
    ''' Test against server specified in environment variable BT_HOST
        and BT_API or MyGene.info production server V3 by default '''

    host = os.getenv("BT_HOST", "http://mygene.info")
    api = os.getenv("BT_API", "/v3")

    def test_01(self):
        ''' KWARGS CTRL Format Json '''
        self.query(q='__all__', size='1')

    def test_02(self):
        ''' KWARGS CTRL Format Yaml '''
        res = self.request('query?q=__all__&size=1&format=yaml').text
        assert res.startswith('max_score:')

    def test_03(self):
        ''' KWARGS CTRL Format Html '''
        res = self.request('query?q=__all__&size=1&format=html').text
        assert '<html>' in res

    def test_04(self):
        ''' KWARGS CTRL Format Msgpack '''
        res = self.request('query?q=__all__&size=1&format=msgpack').content
        self.msgpack_ok(res)

    def test_11(self):
        ''' HANDLE Unmatched Quotes'''
        # Sentry
        # Issue 529121368
        # Event 922fc99638cb4987bccbfd30c914ff03
        _q = 'query?q=c("ZNF398", "U2AF...'
        self.request(_q, expect_status=400)


if __name__ == '__main__':
    main(defaultTest='__main__.QueryTests', argv=['', '-v'])



from biothings.web.api.helper import BaseHandler, BiothingParameterTypeError
from biothings.utils.common import dotdict, is_str
import re
import logging

class BaseESRequestHandler(BaseHandler):
    ''' Parent class of all Elasticsearch-based Request handlers, subclass of `BaseHandler`_. 
    Contains handling for Elasticsearch-specific query params (like ``fields``, ``size``, etc) '''
    # override these in child class
    control_kwargs = {}
    es_kwargs = {}
    esqb_kwargs = {}
    transform_kwargs = {}

    def initialize(self, web_settings):
        ''' Tornado reqeust handler initialization.  Initializations common to all 
        Elasticsearch-specific request handlers go here.'''
        super(BaseESRequestHandler, self).initialize(web_settings)

    def _return_data_and_track(self, data, ga_event_data={}, rawquery=False, status_code=200, _format='json'):
        ''' Small function to return a chunk of data and send a google analytics tracking request.'''
        if rawquery:
            self.return_raw_query_json(data, status_code=status_code, _format=_format)
        else:
            self.return_object(data, status_code=status_code, _format=_format)
        self.ga_track(event=self.ga_event_object(ga_event_data))
        self.self_track(data=self.ga_event_object_ret)
        return

    def return_raw_query_json(self, query, status_code=200, _format='json'):
        '''Return valid JSON if `rawquery` option is selected.
        This is necessary as queries can span multiple lines (POST)'''
        _ret = query.get('body', {'GET': query.get('bid')})
        if is_str(_ret) and len(_ret.split('\n')) > 1:
            self.return_object({'body': _ret}, status_code=status_code, _format=_format)
        else:
            self.return_object(_ret, status_code=status_code, _format=_format)

    def _should_sanitize(self, param, kwargs):
        return ((param in kwargs) and (param in self.kwarg_settings))

    def _sanitize_source_param(self, kwargs):
        if self._should_sanitize('_source', kwargs):
            if len(kwargs['_source']) == 0:
                kwargs.pop('_source')
            elif len(kwargs['_source']) == 1 and kwargs['_source'][0].lower() == 'all':
                kwargs['_source'] = True
        return kwargs

    def _sanitize_sort_param(self, kwargs):
        if self._should_sanitize('sort', kwargs):
            dirty_sort = kwargs['sort']
            kwargs['sort'] = [{field[1:]:{"order":"desc"}} if field.startswith('-') 
                else {field:{"order":"asc"}} for field in dirty_sort]
        return kwargs

    def _sanitize_size_param(self, kwargs):
        # cap size
        if self._should_sanitize('size', kwargs):
            kwargs['size'] = kwargs['size'] if (kwargs['size'] < 
                self.web_settings.ES_SIZE_CAP) else self.web_settings.ES_SIZE_CAP
        return kwargs

    def _parse_nested_aggs(self, field, facet_size):
        ''' parse nested aggs '''
        if re.search(r'\(.*\)', field):
            return {re.sub(r'\(.*\)', '', field): {'terms': {"field": re.sub(r'\(.*\)', '', field), 'size': facet_size}, "aggs": self._parse_nested_aggs(re.sub(r'^.*?\(','', field)[:-1], facet_size)}}
        else:
            return {field: {'terms': {"field": field, 'size': facet_size}}}

    def _sanitize_aggs_param(self, kwargs):
        if self._should_sanitize('aggs', kwargs):
            kwargs['facet_size'] = kwargs.get('facet_size', self.kwarg_settings['facet_size']['default']) 
            if (kwargs['facet_size'] > self.kwarg_settings['facet_size'].get('max', 1000)):
                 kwargs['facet_size'] = self.kwarg_settings['facet_size'].get('max', 1000)
            aggs_dict = {}
            for field in kwargs['aggs']:
                if self.web_settings.ALLOW_NESTED_AGGS and re.search(r'\(.*\)', field):
                    aggs_dict.update(self._parse_nested_aggs(field, kwargs['facet_size']))
                else:
                    aggs_dict.update({field: {"terms": {"field": field, "size": kwargs["facet_size"]}}})
            kwargs['aggs'] = aggs_dict
        return kwargs

    def _sanitize_from_param(self, kwargs):
        # cap from
        if self._should_sanitize('from', kwargs) and kwargs['from'] > self.web_settings.ES_RESULT_WINDOW_SIZE_CAP:
            raise BiothingParameterTypeError('"from" parameter cannot be greater than {0}'.format(self.web_settings.ES_RESULT_WINDOW_SIZE_CAP))
        return kwargs

    def get_cleaned_options(self, kwargs):
        ''' Separate URL keyword arguments into their functional category.
        Incoming kwargs are separated into one of 4 categories, depending on how the argument controls the
        pipeline:

        * ``control_kwargs`` - These are arguments that control the pipeline execution (**raw**, **rawquery**, etc)
        * ``es_kwargs`` - These are arguments that get passed directly to the Elasticsearch client during query
        * ``esqb_kwargs`` - These are arguments that go to the Elasticsearch query builder (**fields**, **size**, etc)
        * ``transform_kwargs`` - These are arguments that go to the Elasticsearch result transformer (**jsonld**, **dotfield**, etc)'''
        options = dotdict()

        # split kwargs into one (or more) of 4 categories:
        #   * control_kwargs:  kwargs that control aspects of the handler's pipeline (e.g. raw, rawquery)
        #   * es_kwargs: kwargs that go directly to the ES query (e.g. fields, size, ...)
        #   * esqb_kwargs: kwargs that go directly to the ESQueryBuilder instance
        #   * transform_kwargs: kwargs that go directly to the response transformer (e.g. jsonld, dotfield)
        for kwarg_category in ["control_kwargs", "es_kwargs", "esqb_kwargs", "transform_kwargs"]:
            options.setdefault(kwarg_category, dotdict())
            for option, settings in getattr(self, kwarg_category, {}).items():
                if option in kwargs or settings.get('default', None) is not None:
                    options.get(kwarg_category).setdefault(option, kwargs.get(option, settings['default']))
        # do userquery kwargs
        if options.esqb_kwargs.userquery:
            for k in kwargs.keys():
                if re.match(self.web_settings.USERQUERY_KWARG_REGEX, k):
                    options['esqb_kwargs'].setdefault('userquery_kwargs', dotdict())
                    options['esqb_kwargs']['userquery_kwargs'][self.web_settings.USERQUERY_KWARG_TRANSFORM(k)] = kwargs.get(k)
        return options

    def _sanitize_params(self, kwargs):
        kwargs = super(BaseESRequestHandler, self)._sanitize_params(kwargs)
        kwargs = self._sanitize_source_param(kwargs)
        kwargs = self._sanitize_size_param(kwargs)
        kwargs = self._sanitize_sort_param(kwargs)
        kwargs = self._sanitize_aggs_param(kwargs)
        kwargs = self._sanitize_from_param(kwargs)
        return kwargs
    
    def _get_es_index(self, options):
        ''' Override to change query index for this request. '''
        return self.web_settings.ES_INDEX

    def _get_es_doc_type(self, options):
        ''' Override to change doc_type for this request. '''
        return self.web_settings.ES_DOC_TYPE

    def _pre_query_builder_GET_hook(self, options):
        ''' Override me. '''
        return options

    def _pre_query_GET_hook(self, options, query):
        ''' Override me. '''
        return query

    def _pre_transform_GET_hook(self, options, res):
        ''' Override me. '''
        return res

    def _pre_finish_GET_hook(self, options, res):
        ''' Override me. '''
        return res
    
    def _pre_query_builder_POST_hook(self, options):
        ''' Override me. '''
        return options

    def _pre_query_POST_hook(self, options, query):
        ''' Override me. '''
        return query

    def _pre_transform_POST_hook(self, options, res):
        ''' Override me. '''
        return res

    def _pre_finish_POST_hook(self, options, res):
        ''' Override me. '''
        return res



'''
    Biothings Test Helper
'''
import json
import re
import sys
import unittest
from difflib import Differ
from functools import partial, wraps
from urllib.parse import urlparse

import requests
from nose import SkipTest
from nose.tools import eq_
from tornado.ioloop import IOLoop
from tornado.testing import AsyncHTTPTestCase
from tornado.web import Application

from biothings.web.settings import BiothingESWebSettings


def equal(type_a, value_a, type_b, value_b):
    ''' Equality assertion with helpful diff messages '''

    def split(string):
        chars_per_line = 68
        return [string[i:i+chars_per_line]+'\n'
                for i in range(0, len(string), chars_per_line)]

    if not value_a == value_b:
        if isinstance(value_a, set) and isinstance(value_b, set):
            print('Objects in', type_a, 'only:')
            print(value_a - value_b)
            print('Objects in', type_b, 'only:')
            print(value_b - value_a)
        else:
            lines1 = split(str(value_a))
            lines2 = split(str(value_b))
            differ = Differ()
            result = list(differ.compare(lines1, lines2))
            start_index = None  # inclusive
            end_index = None  # exclusive
            for index, result_line in enumerate(result):
                if result_line.startswith('  '):  # common lines
                    if start_index:
                        end_index = index
                        break
                    continue
                if not start_index:
                    start_index = index
                else:
                    end_index = index

            if end_index - start_index > 8:  # show 2 mismatch lines max
                end_index = start_index + 8
            else:
                end_index += 3  # show context unless diff is too long
            start_index -= 3  # show context
            if start_index < 0:
                start_index = 0
            if end_index > len(result):
                end_index = len(result)
            result[end_index-1] = result[end_index-1][:-1]  # remove trailing newline
            sys.stdout.writelines(result[start_index:end_index])
        raise AssertionError(type_a + ' != ' + type_b)


class TornadoTestServerMixin(AsyncHTTPTestCase):
    '''
        Starts a tornado server to run tests on
        Must appear first in subclass's inheritance list
        May override with customzied 'settings'
    '''

    host = ''

    def __new__(cls, *args, **kwargs):
        if not getattr(cls, 'settings', None):
            cls.settings = BiothingESWebSettings(config='config')
        return super(TornadoTestServerMixin, cls).__new__(cls)

    # override
    def get_new_ioloop(self):
        return IOLoop.current()

    # override
    def get_app(self):
        app_list = self.settings.generate_app_list()
        static_path = self.settings.STATIC_PATH
        if getattr(self.settings, 'COOKIE_SECRET', None):
            return Application(app_list, static_path=static_path,
                               cookie_secret=self.settings.COOKIE_SECRET)
        return Application(app_list, static_path=static_path)

    # override
    def request(self, *args, **kwargs):
        ''' Use requests.py instead of the built-in client
            Override to make the requests non-blocking
            param: path: network path to make request to
            param: method: ('GET') http request method
            param: expect_status: (200) check status code
        '''

        partial_func = partial(super(TornadoTestServerMixin, self).request, **kwargs)

        async def call_blocking():
            return await self.io_loop.run_in_executor(None, partial_func, *args)

        return self.io_loop.run_sync(call_blocking)


class BiothingsTestCase(unittest.TestCase):
    '''
        Contains common functions to help facilitate testing.
        Assumes that .host .api are set in the subclass
    '''

    # override in subclass
    host = None  # no trailing slash '/'
    api = None  # do not include host, starts with '/'

    #############################################################
    # Helper functions                                          #
    #############################################################

    # HTTP Requests

    def request(self, path, method="GET", expect_status=200, **kwargs):
        '''
           Prefix path with api endpoint unless it specifies an absolute
           or relative URL. Make a request according to the method specified.
           Validate response status code and return the response object. '''

        assert self.api is not None  # allow empty string
        if not path:
            url = self.get_url('/')
        elif path.lower().startswith(("http://", "https://")):
            url = path
        elif path.startswith('/'):
            url = self.get_url(path)
        else:
            url = self.get_url(self.api + '/' + path)
        res = requests.request(method, url, **kwargs)
        eq_(res.status_code, expect_status)
        return res

    def get_url(self, path):
        ''' Takes a relative path starting with '/'
            Append it to host and return the full url
            Need override for local tornado tests '''

        assert path.startswith('/')
        return self.host + path

    def query(self, method='GET', endpoint='query', expect_hits=True, **kwargs):
        ''' Make a query and assert positive hits by default.
            Assert zero hit when expect_hits is set to False. '''

        if method == 'GET':
            dic = self.request(endpoint, params=kwargs).json()
            if expect_hits:
                assert dic.get('hits', []), "No Hits"
            else:
                assert dic.get('hits', None) == [], f"Get {dic.get('hits')} instead."
            return dic

        if method == 'POST':
            lst = self.request(endpoint, method=method, data=kwargs).json()
            hits = False
            for item in lst:
                if "_id" in item:
                    hits = True
                    break
            if expect_hits:
                assert hits
            else:
                assert not hits
            return lst

        raise ValueError(f'Query method {method} is not supported.')

    @staticmethod
    def parse_url(url, option):
        ''' Parse the url string to see if an option is specified,
            if so return the option value, if not return empty string '''
        options = urlparse(url).query.split('&')
        for opt in options:
            if opt.split('=')[0] == option:
                return opt.split('=')[1]
        return ''

    # Data Formats

    @staticmethod
    def truncate(string, limit):
        ''' Truncate a long string with a trailing ellipsis '''
        if len(string) <= limit:
            return string
        return string[:limit] + '...'

    @classmethod
    def msgpack_ok(cls, packed_bytes, checkerror=True):
        ''' Load msgpack into a dict '''
        try:
            import msgpack
        except ImportError:
            raise SkipTest('Msgpack is not installed.')
        try:
            dic = msgpack.unpackb(packed_bytes)
        except:  # pylint: disable=bare-except
            assert False, 'Not a valid Msgpack binary.'
        if checkerror:
            assert not (isinstance(dic, dict)
                        and 'error' in dic), cls.truncate(str(dic), 100)
        return dic

    @staticmethod
    def convert_msgpack(obj):
        '''
            obj is a msgpack decoded dict (strings are still byte objects).
            Traverse through obj and decode all bytes into python strings.  Return the result.
        '''
        def convert_str(k):
            if isinstance(k, bytes):
                return k.decode('utf-8')
            return k

        def traverse(obj):
            if isinstance(obj, list):
                return [traverse(i) for i in obj]
            if isinstance(obj, dict):
                return {convert_str(tk): traverse(tv) for (tk, tv) in obj.items()}
            return convert_str(obj)

        return traverse(obj)

    def check_jsonld(self, jsld, jsonld_context):
        '''
            Traverse through jsld and assert that JSON-LD contexts are inserted and then removed
            should leave jsld a context-less JSON object...(no guarantee, see below)
        '''
        # Currently only tests that contexts in context file are in the object, and
        # removes them.  Maybe should add an else to the if not k: clause, and test that no
        # @context are in objects where they shouldn't be, to be really complete
        def traverse(jsld, k):
            # valid jsonld context?
            if isinstance(jsld, list):
                return [traverse(i, k) for i in jsld]
            if isinstance(jsld, dict):
                if not k:
                    # Root
                    self.assertIn(
                        '@context', jsld, "JSON-LD context not found in root.  Expected: {}"
                        .format(jsonld_context['root']))
                    self.assertDictEqual(
                        jsonld_context['root']['@context'], jsld['@context'])
                    del jsld['@context']
                    return {tk: traverse(tv, tk) for (tk, tv) in jsld.items()}
                if k in jsonld_context['root']['@context'] and k not in jsonld_context:
                    # No context, but defined in root context
                    return {tk: traverse(tv, k + '/' + tk) for (tk, tv) in jsld.items()}
                if k in jsonld_context:
                    # Context inserted, test it, and remove it
                    self.assertIn(
                        '@context', jsld, "JSON-LD context not found in {}.  Expected: {}"
                        .format(k, jsonld_context[k]))
                    self.assertDictEqual(
                        jsonld_context[k]['@context'], jsld['@context'])
                    del jsld['@context']
                    return {tk: traverse(tv, k + '/' + tk) for (tk, tv) in jsld.items()}
            return jsld
        if 'root' not in jsonld_context:
            return jsld  # can't check anything
        jsonld_context = jsonld_context  # maybe not needed...
        traverse(jsld, '')

    def extract_results_from_callback(self, base_url):
        '''
            Given a base query (with callback), the corresponding escaped GET url,
            test the response and return the JSON object inside
        '''
        res_text = self.request(base_url).text()
        callback = self.parse_url(base_url, 'callback')
        res_text = re.sub('\n', '', res_text)
        pattern = r'^(?P<callback>' + callback + r'\()(?P<res>\{.*\})\)$'
        match = re.search(pattern, res_text)
        assert match, 'JSONP object malformed'
        result = match.groupdict()
        # get the json object out of the callback so we can test it
        return json.loads(result['res'])

    @staticmethod
    def check_fields(res, res_all, fields, additionals=None):
        ''' Asserts that the fields in a response corresponds to the "fields" parameter in request
                :param res: the response to a test request
                :param res_all: the response to a "field = all" reference request
                :param list fields: fields specified in the test request e.g. ['cadd.gene']
                :param list additionals: additionally supported fields besides
                    typical biothing fields  eg. _license for cadd in myvariant
        '''
        def flatten_dict(json_obj, prefix, result):
            ''' Flattens the keys of a multi-level json object using dot notation
            returns a list of all dotted keys in the object, leaving a . in front :) '''
            if isinstance(json_obj, list):
                for obj in json_obj:
                    flatten_dict(obj, prefix, result)
            elif isinstance(json_obj, dict):
                for key in json_obj.keys():
                    result[prefix + '.' + key] = 0
                    flatten_dict(json_obj[key], prefix + '.' + key, result)

        def expand_fields(requested_fields, all_fields):
            ''' Returns the list of all possible keys in a valid response object,
            based on the users fields input, e.g.: cadd.gene means that cadd and cadd.gene
            should be in the output (if available), similarly, if clinvar is not in the total
            object, then it should not be in the output (even if the user specifies it).
            '''
            # find all possible fields from the request
            possible_fields = []
            if requested_fields[0] == 'all':
                return all_fields
            for field in requested_fields:
                possible_fields += [s for s in all_fields if s
                                    == field or s.startswith(field + '.')]
            # Go through and add parent nodes, which must be in the object....
            pfs = set(possible_fields)
            for fld in possible_fields:
                key = ''
                for path in fld.split('.'):
                    key = key + '.' + path if key else path
                    pfs.add(key)
            return list(pfs)

        # extract fields from "fields=all" response
        fields_all_flattened = {}
        flatten_dict(res_all, '', fields_all_flattened)
        # produce comparable keyword list
        fields_all_expanded = expand_fields(
            fields, [x.lstrip('.') for x in fields_all_flattened])
        # add root fields (_id, _version, query, _score)
        if not additionals:
            additionals = []
        fields_all_additional = ['_id', '_version',
                                 'query', '_score'] + additionals
        # combine comparable keyword lists into a set
        fields_all = set(fields_all_expanded + fields_all_additional)

        # extract fields from response to test
        fields_res_flattened = {}
        flatten_dict(res, '', fields_res_flattened)
        # produce comparable keyword set
        fields_res = {x.lstrip('.') for x in fields_res_flattened}

        # asserts no unspecified fields are returned
        assert fields_res.issubset(fields_all), \
            "The returned keys of object {} have extra keys than expected, the offending keys \
            are: {}".format(res['_id'], fields_res.difference(fields_all))


#############################################
# Decorators for easy parameterized testing
#############################################


PATTR = '%values'


def parameterized(cls):
    ''' Class decorator to add all parameterized test cases to class '''
    def feed(func, new_test_name, val):
        ''' produce a new test function '''
        @wraps(func)
        def wrapper(self):
            return func(self, val)
        wrapper.__name__ = new_test_name
        if func.__doc__:
            wrapper.__doc__ = func.__doc__
        return wrapper

    for name, method in list(cls.__dict__.items()):
        if hasattr(method, PATTR):
            for i, val in enumerate(getattr(method, PATTR)[0]):
                new_test_name = "{}_{}".format(name, i + 1)
                setattr(cls, new_test_name, feed(method, new_test_name, val))
            delattr(cls, name)
    return cls


def parameters(*values):
    ''' Method decorator to add parameter lists to template test methods '''
    def wrapper(func):
        setattr(func, PATTR, values)
        return func
    return wrapper



import logging
import json
import os
from biothings.utils.common import is_seq
from biothings.utils.web.userquery import get_userquery, get_userfilter
try:
    from re import fullmatch as match
except ImportError:
    from re import match

class ESQueries(object):
    ''' A very simple class to object-ize Elasticsearch Query DSL.
    This should be replaced by the official `Elasticsearch equivalent <https://pypi.python.org/pypi/elasticsearch-dsl>`_.
    Also contains a simple JSON validator after generating all queries (dump to string and re-read) '''
    def __init__(self, es_kwargs={}):
        self.es_kwargs = es_kwargs

    def _validate_json(self, d):
        ''' dumps the dict to string and reads back in as json - should validate the json structure? '''
        try:
            return json.loads(json.dumps(d))
        except:
            raise ValueError('Could not validate query "{}"'.format(d))

    def _es_query_template(self, query_type=None, query_kwargs=None, raw_query={}):
        _ret = raw_query if raw_query else {"query":{query_type: query_kwargs}}
        _ret.update(self.es_kwargs)
        return self._validate_json(_ret)

    def multi_match(self, query_kwargs):
        ''' Given ``query_kwargs``, validate and return a **multi_match** query. '''
        return self._es_query_template(query_type="multi_match", query_kwargs=query_kwargs)

    def match(self, query_kwargs):
        ''' Given ``query_kwargs``, validate and return a **match** query. '''
        return self._es_query_template(query_type="match", query_kwargs=query_kwargs)

    def match_all(self, query_kwargs):
        ''' Given ``query_kwargs``, validate and return a **match_all** query. '''
        return self._es_query_template(query_type="match_all", query_kwargs=query_kwargs)

    def function_score(self, query_kwargs):
        ''' Given ``query_kwargs``, validate and return a **function score** query. '''
        return self._es_query_template(query_type="function_score", query_kwargs=query_kwargs)

    def query_string(self, query_kwargs):
        ''' Given ``query_kwargs``, validate and return a **query_string** query. '''
        return self._es_query_template(query_type="query_string", query_kwargs=query_kwargs)

    def bool(self, query_kwargs):
        ''' Given ``query_kwargs``, validate and return a **bool** query. '''
        return self._es_query_template(query_type="bool", query_kwargs=query_kwargs)

    def raw_query(self, raw_query):
        ''' Given ``query_kwargs``, validate and return a *raw* query (queries that don't fit the same query_template). '''
        return self._es_query_template(raw_query=raw_query)

class ESQueryBuilder(object):
    ''' Class to return the correct query given the request endpoint, method, and URL params.

    :param index: The Elasticsearch index to run the query on 
    :param doc_type: The Elasticsearch document type of the query
    :param options: Options from the URL string relevant to query building 
    :param es_options: Options for Elasticsearch query stage 
    :param scroll_options: Options for scroll requests
    :param regex_list: A list of (regex, scope) tuples for annotation lookup
    :param userquery_dir: The directory containing user queries for this app
    :param default_scopes: A list representing the default Elasticsearch query scope(s) for this query'''
    def __init__(self, index, doc_type, options, es_options, scroll_options={}, 
                       userquery_dir='', regex_list=[], default_scopes=['_id'], allow_random_query=False):
        self.index = index
        self.doc_type = doc_type
        self.options = options
        self.es_options = es_options
        self.scroll_options = scroll_options
        self.regex_list = regex_list
        self.userquery_dir = userquery_dir
        self.default_scopes = default_scopes
        self.queries = ESQueries(es_options)
        self.allow_random_query = allow_random_query

    def _return_query_kwargs(self, query_kwargs):
        _kwargs = {"index": self.index, "doc_type": self.doc_type}
        _kwargs.update(query_kwargs)
        return _kwargs 

    def _get_term_scope(self, term):
        _scopes = None
        _term = term
        for (regex, scope) in self.regex_list:
            r = match(regex, term)
            if r:
                if r.groupdict() and 'search_term' in r.groupdict() and r.groupdict()['search_term']:
                    _scopes = scope
                    _term = r.groupdict()['search_term']
                    break
                else:
                    _scopes = scope
                    break
        return _term, _scopes
    
    def _build_single_query(self, term, scopes=None):
        scopes = scopes or self.default_scopes
        if len(scopes) == 1:
            return self.queries.match({scopes[0]:{"query": "{}".format(term), "operator": "and"}})
        else:
            return self.queries.multi_match({"query":"{}".format(term), "fields":scopes, "operator":"and"})

    def _build_multiple_query(self, terms, scopes=None):
        _q = []
        _infer_scope = True if not scopes else False
        for term in terms:
            _term = term
            if _infer_scope:
                _term, scopes = self._get_term_scope(term)
            _q.extend(['{}', json.dumps(self._build_single_query(_term, scopes=scopes))])
        return self._return_query_kwargs({'body': '\n'.join(_q)})

    def _default_query(self, q):
        ''' Override me '''
        return ESQueries().query_string({"query": q})

    def _is_match_all(self, q):
        return (q == '__all__')

    def _is_random_query(self, q):
        return (q == '__any__')

    def _extra_query_types(self, q):
        ''' Override me '''
        return {}

    def _match_all(self, q):
        ''' Override me '''
        return ESQueries().match_all({})
    
    def _random_query(self, q):
        return ESQueries().function_score({"random_score": {}})

    def _is_user_query(self, text_file='query.txt'):
        try:
            query_dir = os.path.join(os.path.abspath(self.userquery_dir), self.options.userquery)
            return (os.path.exists(query_dir) and (os.path.isdir(query_dir)) and 
                    (os.path.exists(os.path.join(query_dir, text_file))))
        except Exception:
            return False
    
    def _user_query(self, q):
        _args = {'q': q}
        if self.options.userquery_kwargs:
            _args.update(self.options.userquery_kwargs)
        _ret = json.loads(get_userquery(os.path.abspath(self.userquery_dir), 
                            self.options.userquery).format(**_args))
        return ESQueries().raw_query(_ret)

    def _user_query_filter(self):
        return json.loads(get_userfilter(os.path.abspath(self.userquery_dir), self.options.userquery))

    def _get_query_filters(self):
        _filter = []
        if self._is_user_query(text_file='filter.txt'):
            _filter.append(self._user_query_filter())
        return _filter

    def _get_missing_filters(self):
        return []

    def get_query_filters(self):
        ''' Override me to add more query filters '''
        return self._get_query_filters()

    def get_missing_filters(self):
        ''' Override me to add more must_not filters '''
        return self._get_missing_filters()

    def add_extra_filters(self, q):
        ''' Override me to add more filters '''
        return q

    def add_query_filters(self, _query):
        ''' Given a query, add any other filters '''
        filters = self.get_query_filters()
        missing = self.get_missing_filters()
        if not filters and not missing:
            return _query

        #add filters as filtered query
        #this will apply to facet counts
        _query = {
            'bool': {
                'must': _query.get('query', _query)
            }
        }
        if filters:
            _query['bool']['filter'] = filters
        if missing:
            _query['bool']['must_not'] = missing

        return _query
    
    def _scroll(self, scroll_id):
        return {'body': {'scroll_id': scroll_id}, 'scroll': self.scroll_options.get('scroll', '1m')}

    def _annotation_GET_query(self, bid):
        _bid, _scopes = self._get_term_scope(bid)
        if _scopes:
            return self._return_query_kwargs({'body': self._build_single_query(_bid, scopes=_scopes)})
        else:
            # go to es.get
            _get_kwargs = {'id': bid}
            _get_kwargs.update(self.es_options)
            return self._return_query_kwargs(_get_kwargs)
    
    def _annotation_POST_query(self, bids):
        return self._build_multiple_query(terms=bids)

    def _query_GET_query(self, q):
        if self._is_user_query():
            _query = self._user_query(q)
        elif self._is_match_all(q):
            _query = self._match_all(q)
        elif self._is_random_query(q) and self.allow_random_query:
            _query = self._random_query(q)
        else:
            _query = self._extra_query_types(q)

        if not _query:
            _query = self._default_query(q)

        _query = self.add_query_filters(_query)
        _query = self.add_extra_filters(_query)

        _query = self.queries.raw_query(_query)

        _ret = self._return_query_kwargs({'body': _query})

        if self.options.fetch_all:
            _ret['body'].pop('sort', None)  # don't allow sorting for fetch all, defeats the purpose
            _ret['body'].pop('size', None)
            _ret.update(self.scroll_options)
        return _ret

    def _query_POST_query(self, qs, scopes):
        return self._build_multiple_query(terms=qs, scopes=scopes)

    def _metadata_query(self):
        return self._return_query_kwargs({})

    def annotation_GET_query(self, bid):
        ''' Return an annotation lookup GET query for this ``bid``. 

        :param bid: Biothing ID, used to lookup the annotation'''
        return self._annotation_GET_query(bid)
    
    def annotation_POST_query(self, bids):
        ''' Return an annotation lookup POST query for these ``bids``. 

        :param bids: Biothing IDs, used to lookup the annotations'''
        return self._annotation_POST_query(bids)

    def query_GET_query(self, q):
        ''' Return a query endpoint GET query for this query string ``q``. 

        :param q: query string specifying the query'''
        return self._query_GET_query(q)

    def query_POST_query(self, qs, scopes):
        ''' Return query endpoint POST queries for these query strings ``qs``. 

        :param qs: Query strings to query
        :param scopes: Scope of query strings ``qs``'''
        return self._query_POST_query(qs, scopes)

    def metadata_query(self):
        ''' Return a metadata query ''' 
        return self._metadata_query()

    def scroll(self, scroll_id):
        ''' Return the next batch of results from a *scroll* query. 

        :param scroll_id: ID of the batch to yield hits from '''
        return self._scroll(scroll_id)
